
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysBase.cpp
**
** DESCRIPTION: This class mainly manage the file system used by the load
**              generator.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientFileSysBase::CClientFileSysBase()
**     CClientFileSysBase::~CClientFileSysBase()
**
**
****************************************************************************/


#include <stdio.h>
#include <string.h>
#include <time.h>
#ifndef WIN32
#include <ctype.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/mount.h>
#include <sys/mntent.h>
#include <sys/mnttab.h>
#include <sys/statvfs.h>
#include <unistd.h>
#include <errno.h>
#endif
#include "Common.h"
#include "ClientFileSysBase.h"
#include "SafeFileOperation.h"
#include "Packet.h"
#include "SocketClient.h"
#include "NetXDT.h"
#include "Pumper.h"
#include "ClientFileSysRefPool.h"

#ifndef WIN32
extern CLocalPathManager theLocalPathManager;
#endif

CClientFileSysBase::CClientFileSysBase(char *servername, char *servernetif, char *serverpath, char *localpath)
{
    m_noac = TRUE;
    m_mounted = FALSE;

    strcpy(m_ServerName, servername);
    strcpy(m_ServerNetIF, servernetif);
    strcpy(m_ServerPath, serverpath);
    strcpy(m_LocalPath, localpath);

    m_SubDirNum = 0;

    m_IONum = NULL;
    m_IOSize = NULL;
    m_FileNum = NULL;
    m_FileMode = NULL;
    m_DirNum = NULL;
    m_DirMode = NULL;
    m_SymlinkNum = NULL;
    m_SymlinkMode = NULL;

    InitMutex(m_break_lock);
    m_break_value = FALSE;
}

CClientFileSysBase::CClientFileSysBase(char *servername, char *servernetif, char *serverpath, char *localpath,
                                       int totalsize, int totalfnodenum,
                                       int subdirnum, int ionum, int filenum, int dirnum, int symlinknum)
{
    m_noac = TRUE;
    m_mounted = FALSE;

    strcpy(m_ServerName, servername);
    strcpy(m_ServerNetIF, servernetif);
    strcpy(m_ServerPath, serverpath);
    strcpy(m_LocalPath, localpath);
    m_TotalSize = totalsize;
    m_TotalFNodeNum = totalfnodenum;
    m_SubDirNum = subdirnum;
    m_TotalIONum = ionum;
    m_TotalFileNum = filenum;
    m_TotalDirNum = dirnum;
    m_TotalSymlinkNum = symlinknum;

    m_SubDirNum = 0;

    m_IONum = NULL;
    m_IOSize = NULL;
    m_FileNum = NULL;
    m_FileMode = NULL;
    m_DirNum = NULL;
    m_DirMode = NULL;
    m_SymlinkNum = NULL;
    m_SymlinkMode = NULL;

    InitMutex(m_break_lock);
    m_break_value = FALSE;
}

CClientFileSysBase::~CClientFileSysBase()
{
    FreeAllSpace();
#ifndef WIN32
    if (IsMounted())
    {
        if (Unmount())
            theLocalPathManager.FreeLocalPath(m_LocalPath);
    }
#else
    Unmount();

    if (m_break_lock != NULL)
        CloseHandle(m_break_lock);
#endif
}

void CClientFileSysBase::FreeAllSpace()
{
    for (int i=0; i<m_SubDirNum; i++)
    {
        delete [] m_IOSize[i];
        delete [] m_FileMode[i];
        delete [] m_DirMode[i];
        delete [] m_SymlinkMode[i];
    }

    if (m_IONum != NULL)
        delete [] m_IONum;
    if (m_FileNum != NULL)
        delete [] m_FileNum;
    if (m_DirNum != NULL)
        delete [] m_DirNum;
    if (m_SymlinkNum != NULL)
        delete [] m_SymlinkNum;

    if (m_IOSize != NULL)
        delete [] m_IOSize;
    if (m_FileMode != NULL)
        delete [] m_FileMode;
    if (m_DirMode != NULL)
        delete [] m_DirMode;
    if (m_SymlinkMode != NULL)
        delete [] m_SymlinkMode;

    m_SubDirNum = 0;
    m_IONum = NULL;
    m_IOSize = NULL;
    m_FileNum = NULL;
    m_FileMode = NULL;
    m_DirNum = NULL;
    m_DirMode = NULL;
    m_SymlinkNum = NULL;
    m_SymlinkMode = NULL;
}

char *CClientFileSysBase::GetServerName()   { return m_ServerName; }
char *CClientFileSysBase::GetServerNetIF()  { return m_ServerNetIF; }
char *CClientFileSysBase::GetServerPath()   { return m_ServerPath; }
char *CClientFileSysBase::GetLocalPath()    { return m_LocalPath; }

void CClientFileSysBase::SetBreak(BOOL break_value)
{
    LockMutex(m_break_lock);
    m_break_value = break_value;
    UnlockMutex(m_break_lock);
}

BOOL CClientFileSysBase::IsBreak()
{
    LockMutex(m_break_lock);
    int isBreak = m_break_value;
    UnlockMutex(m_break_lock);
    return isBreak;
}

void CClientFileSysBase::SetFileSizeDist(SFileSizeDist *dist)
{
    m_IODist.num = dist->num;
    for (int i=0; i<dist->num; i++)
    {
        m_IODist.item[i].pcnt = dist->item[i].pcnt;
        m_IODist.item[i].size = dist->item[i].size;
    }
}

void CClientFileSysBase::GetPathDirName(char *name)
{
#ifndef WIN32
    int c = '/';
#else
    int c = '\\';
#endif
    char *p = strrchr(name, c);
    if (p != NULL)
        *p = 0;
}

void CClientFileSysBase::GetSubName(char *name, int subIndex)
{
#ifndef WIN32
    sprintf(name, "%s/%04d", m_LocalPath, subIndex);
#else
    sprintf(name, "%s\\%04d", m_LocalPath, subIndex);
#endif
}

void CClientFileSysBase::GetIOFileName(char *name, int subIndex, int fileIndex)
{
#ifndef WIN32
    sprintf(name, "%s/%04d/F%06d", m_LocalPath, subIndex, fileIndex);
#else
    sprintf(name, "%s\\%04d\\F%06d", m_LocalPath, subIndex, fileIndex);
#endif
}

void CClientFileSysBase::GetNonIOFileName(char *name, int subIndex, int fileIndex)
{
#ifndef WIN32
    sprintf(name, "%s/%04d/G%06d", m_LocalPath, subIndex, fileIndex);
#else
    sprintf(name, "%s\\%04d\\G%06d", m_LocalPath, subIndex, fileIndex);
#endif
}

void CClientFileSysBase::GetDirName(char *name, int subIndex, int dirIndex)
{
#ifndef WIN32
    sprintf(name, "%s/%04d/D%06d", m_LocalPath, subIndex, dirIndex);
#else
    sprintf(name, "%s\\%04d\\D%06d", m_LocalPath, subIndex, dirIndex);
#endif
}

void CClientFileSysBase::GetSymlinkName(char *name, int subIndex, int dirIndex)
{
#ifndef WIN32
    sprintf(name, "%s/%04d/L%06d", m_LocalPath, subIndex, dirIndex);
#else
    sprintf(name, "%s\\%04d\\L%06d", m_LocalPath, subIndex, dirIndex);
#endif
}

void CClientFileSysBase::GetSymlinkTargetName(char *name)
{
    /* put a random string into target name */
    int value = GetRandom() % 100;
    if (value < 33)
        strcpy(name, "/DIR1/DIR2/FILE");
    else if (value < 66)
        strcpy(name, "../../DIR/FILE");
    else
        strcpy(name, "DIR1/DIR2/FILE");
}

void CClientFileSysBase::SetCached()
{
    printf("file system is set as cached\n");
    m_noac = FALSE;
}

BOOL CClientFileSysBase::IsMounted()
{
    return m_mounted;
}

BOOL CClientFileSysBase::CheckAndMount()
{
#ifndef WIN32
    int tag = CheckMounted();
    if (tag == 0)
    {
        theLocalPathManager.ApplyLocalPath(m_LocalPath);
        m_mounted = Mount();
    }
    else if (tag == 1)
        m_mounted = TRUE;
    else
    {
        if (Unmount())
            m_mounted = Mount();
        else
            m_mounted = FALSE;
    }
#else
    m_mounted = Mount();
#endif
    return m_mounted;
}

/* return 0:not mounted   1: mounted with right option   2: mounted with wrong option */
int CClientFileSysBase::CheckMounted()
{
#ifdef WIN32
    return 0;
#else
    FILE *fp = fopen("/etc/mnttab", "r");
    if (fp == NULL)
        return 0;

    char special[128], opt[32];
    sprintf(special, "%s:%s", m_ServerNetIF, m_ServerPath);
    strcpy(opt, "noac");
    struct mnttab mp;
    BOOL noac;
    while ( 1 )
    {
        if (getmntent(fp, &mp) != 0)
            break;
        if (strcmp(mp.mnt_special, special) != 0)
            continue;

        fclose(fp);
        strcpy(m_LocalPath, mp.mnt_mountp);
        noac = (hasmntopt(&mp, opt) != NULL);
        if (m_noac)
            return noac ? 1 : 2;
        else
            return noac ? 2 : 1;
    }
    fclose(fp);
    return 0;
#endif
}

BOOL CClientFileSysBase::Mount()
{
#ifdef WIN32
    NETRESOURCE nr;
    DWORD dwResult;

    nr.dwType = RESOURCETYPE_DISK;
    nr.lpLocalName = NULL;
    nr.lpRemoteName = m_LocalPath;
    nr.lpProvider = NULL;

    dwResult = WNetAddConnection2(&nr,  // NETRESOURCE
                (LPSTR) NULL,           // no password
                (LPSTR) NULL,           // logged-in user
                0);                     // withoud profile

    if (dwResult != NO_ERROR)
    { 
        printf("Can't connect to %s: error code is %d\n", m_LocalPath, dwResult);
        return FALSE;
    }

    // Otherwise, report a successful connection.
    //printf("Connected to specified resource.\n");
    return TRUE;

#else
    char *p, *q;
    p = m_LocalPath+1;
    while ((q = strchr(p, '/')) != NULL)
    {
        *q = 0;
        if (CSafeFileOperation::Lookup(m_LocalPath, NULL, NULL) != 1)
            CSafeFileOperation::Mkdir(m_LocalPath);
        *q = '/';
        p = q + 1;
    }
    if (CSafeFileOperation::Lookup(m_LocalPath, NULL, NULL) != 1)
        CSafeFileOperation::Mkdir(m_LocalPath);

    char cmdLine[1024];
    if (m_noac)
        //sprintf(cmdLine, "mount -o noac %s:%s %s >dev/null 2>&1", m_ServerNetIF, m_ServerPath, m_LocalPath);
        sprintf(cmdLine, "mount -o noac,retry=0 %s:%s %s", m_ServerNetIF, m_ServerPath, m_LocalPath);
    else
        //sprintf(cmdLine, "mount %s:%s %s >dev/null 2>&1", m_ServerNetIF, m_ServerPath, m_LocalPath);
        sprintf(cmdLine, "mount -o retry=0 %s:%s %s", m_ServerNetIF, m_ServerPath, m_LocalPath);
    printf("system(%s)\n", cmdLine);
    BOOL succ = (system(cmdLine) == 0);
    if (! succ)
        theLocalPathManager.FreeLocalPath(m_LocalPath);
    return succ;
    /*system(cmdLine);
    int aa = CheckMounted();
    printf("CheckMounted() == %d", aa);
    getchar();
    return aa == 1;*/
#endif
}

BOOL CClientFileSysBase::Unmount()
{
    BOOL bSucc;
#ifdef WIN32
    bSucc = (WNetCancelConnection(m_LocalPath, TRUE) == NO_ERROR);
#else
    //return umount(m_LocalPath) == 0;

    char cmdLine[1024];
    //sprintf(cmdLine, "umount %s >dev/null 2>&1", m_LocalPath);
    sprintf(cmdLine, "umount %s", m_LocalPath);
    bSucc = (system(cmdLine) == 0);
#endif
    if (bSucc)
        m_mounted = FALSE;
    return bSucc;
}

BOOL CClientFileSysBase::CheckWritePermission()
{
    char name[MAX_NAMELEN];
    char file_name[MAX_PATHLEN];

    sprintf(name, "_TEMP_%d", (int)time((time_t*) NULL));
    strcpy(file_name, m_LocalPath);
    strcat(file_name, _STR_PATHSEP);
    strcat(file_name, name);

    return CSafeFileOperation::CheckWritePermission(file_name);
}

char *CClientFileSysBase::GetHostName()
{
    return m_HostName;
}

char *CClientFileSysBase::GetHostIp()
{
    return m_HostIp;
}

void CClientFileSysBase::GetID(int &job_id, int &client_id)
{
    job_id = m_job_id;
    client_id = m_client_id;
}

BOOL CClientFileSysBase::GetUpdateProgress()
{
    return m_bUpdate;
}

void CClientFileSysBase::SetHostName(char *hostname)
{
    strcpy(m_HostName, hostname);
}

void CClientFileSysBase::SetHostIp(char *hostip)
{
    strcpy(m_HostIp, hostip);
}

void CClientFileSysBase::SetID(int job_id, int client_id)
{
    m_job_id = job_id;
    m_client_id = client_id;
}

void CClientFileSysBase::SetUpdateProgress(BOOL bUpdate)
{
    m_bUpdate = bUpdate;
}

void CClientFileSysBase::Rebuild()
{
    FreeAllSpace();

    CalFileNum();

    int i, j, k, dIndex, iIndex;
    char name[MAX_PATHLEN];

    printf("In CClientFileSysBase::Rebuild(): create %d subdir\n", m_SubDirNum);
    /* create sub directories */
    for (i=0; i<m_SubDirNum; i++)
    {
        GetSubName(name, i);
        CSafeFileOperation::Mkdir(name);
    }

    printf("In CClientFileSysBase::Rebuild(): create %d IO files\n", m_TotalIONum);
    /* create IO files */
    m_IONum = new int [m_SubDirNum];
    m_IOSize = new int * [m_SubDirNum];
    int num = m_TotalIONum / m_SubDirNum;
    int remainder = m_TotalIONum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_IONum[i] = (i<remainder) ? num+1 : num;
        m_IOSize[i] = new int [m_IONum[i]];
    }
    for (i=j=k=dIndex=iIndex=0; k<m_TotalIONum; k++)
    {
        GetIOFileName(name, i, j);
        m_IOSize[i][j] = m_IODist.item[iIndex].size * KB;
        CSafeFileOperation::CreateWrite(name, m_IOSize[i][j]);
        dIndex ++;
        if (dIndex == m_IODist.item[iIndex].pcnt)
        {
            iIndex ++;
            if (iIndex == m_IODist.num)
                dIndex = iIndex = 0;
        }
        i ++;
        if (i == m_SubDirNum)
        {
            i = 0;
            j ++;
        }
    }

    printf("In CClientFileSysBase::Rebuild(): create %d NonIO files\n", m_TotalFileNum);
    /* create NonIO files */
    m_FileNum = new int [m_SubDirNum];
    m_FileMode = new ushort * [m_SubDirNum];
    num = m_TotalFileNum / m_SubDirNum;
    remainder = m_TotalFileNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_FileNum[i] = (i<remainder) ? num+1 : num;
        m_FileMode[i] = new ushort [m_FileNum[i]];
    }
    for (i=j=k=0; k<m_TotalFileNum; k++)
    {
        GetNonIOFileName(name, i, j);
        CSafeFileOperation::CreateNull(name, m_FileMode[i][j]);
        i ++;
        if (i == m_SubDirNum)
        {
            i = 0;
            j ++;
        }
    }

    printf("In CClientFileSysBase::Rebuild(): create %d directories\n", m_TotalDirNum);
    /* create directories */
    m_DirNum = new int [m_SubDirNum];
    m_DirMode = new uchar * [m_SubDirNum];
    num = m_TotalDirNum / m_SubDirNum;
    remainder = m_TotalDirNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_DirNum[i] = (i<remainder) ? num+1 : num;
        m_DirMode[i] = new uchar [SIZE_BIT_TO_BYTE(m_DirNum[i])];
    }
    for (i=j=k=0; k<m_TotalDirNum; k++)
    {
        GetDirName(name, i, j);
        if ((j % 8) == 0)
            m_DirMode[i][j/8] = (uchar) 0x01;
        else
            m_DirMode[i][j/8] |= (1 << (j % 8));
        CSafeFileOperation::Mkdir(name);
        i ++;
        if (i == m_SubDirNum)
        {
            i = 0;
            j ++;
        }
    }

    printf("In CClientFileSysBase::Rebuild(): create %d symlinks\n", m_TotalSymlinkNum);
    /* create directories */
    m_SymlinkNum = new int [m_SubDirNum];
    m_SymlinkMode = new uchar * [m_SubDirNum];
    num = m_TotalSymlinkNum / m_SubDirNum;
    remainder = m_TotalSymlinkNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_SymlinkNum[i] = (i<remainder) ? num+1 : num;
        m_SymlinkMode[i] = new uchar [SIZE_BIT_TO_BYTE(m_SymlinkNum[i])];
    }
    char target_name[MAX_PATHLEN];
    for (i=j=k=0; k<m_TotalSymlinkNum; k++)
    {
        GetSymlinkName(name, i, j);
        if ((j % 8) == 0)
            m_SymlinkMode[i][j/8] = (uchar) 0x01;
        else
            m_SymlinkMode[i][j/8] |= (1 << (j % 8));
        GetSymlinkTargetName(target_name);
        CSafeFileOperation::Symlink(name, target_name);
        i ++;
        if (i == m_SubDirNum)
        {
            i = 0;
            j ++;
        }
    }

    printf("In CClientFileSysBase::Rebuild(): finish\n");
}

void CClientFileSysBase::Refresh()
{
    FreeAllSpace();

    CalFileNum();

    if (m_bUpdate)
        UpdateProgress(1, FALSE);
    Refresh_SubDirs();
    if (m_bUpdate)
        UpdateProgress(5, FALSE);

    //Refresh_Content();
    Refresh_Content_Multithread();
}

/* create sub directories */
void CClientFileSysBase::Refresh_SubDirs()
{
    char dir_name[MAX_PATHLEN], name[MAX_PATHLEN], file_name[MAX_PATHLEN];

    printf("In CClientFileSysBase::Refresh_SubDirs()\n");
    printf("        refresh %d subdir\n", m_SubDirNum);

    BOOL *isFound = new BOOL [m_SubDirNum];
    int i, j;
    for (i=0; i<m_SubDirNum; i++)
        isFound[i] = FALSE;

    GetSubName(dir_name, 0);
    GetPathDirName(dir_name);

#ifndef WIN32

    /* open this direcotry */
    DIR *dirp = opendir(dir_name);
    if (dirp == NULL)
    {
        fprintf(stderr, "Make the root direcotry first\n");
        return;
    }

    /* search this directory */
    struct dirent *dp;
    while ((dp = readdir(dirp)) != NULL)
    {
        strcpy(name, dp->d_name);
        if (name[0] == '.')     /* ignore "." && ".." */
            continue;
        j = atoi(name);
        if (j == 0 && (strlen(name) != 4 || !isdigit(name[0])))
            j = -1;
        if (j >= 0 && j < m_SubDirNum)
            isFound[j] = TRUE;
        else if (j > 0)
        {
            strcpy(file_name, dir_name);
            strcat(file_name, "/");
            strcat(file_name, name);
            CSafeFileOperation::RemoveAny(file_name);
        }
    }

    closedir(dirp);

#else

    /* open this direcotry */
    lstrcpy(file_name, dir_name);
    lstrcat(file_name, "\\*.*");
    WIN32_FIND_DATA FileData;
    HANDLE hSearch = FindFirstFile(file_name, &FileData);
    if (hSearch == INVALID_HANDLE_VALUE)
    {
        fprintf(stderr, "Make the root direcotry first\n");
        return;
    }

    /* serarch this directory */
    do 
    {
        strcpy(name, FileData.cFileName);
        if (name[0] != '.')     /* ignore "." && ".." */
        {
            j = atoi(name);
            if (j == 0 && (strlen(name) != 4 || !isdigit(name[0])))
                j = -1;
            if (j >= 0 && j < m_SubDirNum)
                isFound[j] = TRUE;
            else if (j > 0)
            {
                strcpy(file_name, dir_name);
                strcat(file_name, "\\");
                strcat(file_name, name);
                CSafeFileOperation::RemoveAny(file_name);
            }
        }
    } while (FindNextFile(hSearch, &FileData));

    FindClose(hSearch);

#endif

    for (i=0; i<m_SubDirNum; i++)
    {
        if (! isFound[i])
        {
            GetSubName(name, i);
            CSafeFileOperation::Mkdir(name);
        }
    }

    delete [] isFound;
}

/* create IO files, create NonIO files, create dirs, create NonIO files */
void CClientFileSysBase::Refresh_Content()
{
    int i, j, k, dIndex, iIndex, size;
    char dir_name[MAX_PATHLEN], name[MAX_PATHLEN], file_name[MAX_PATHLEN], target_name[MAX_PATHLEN];
    BOOL *foundIO, *foundFile, *foundDir, *foundSymlink, deleteTag;

    printf("In CClientFileSysBase::Refresh_Content()\n");
    printf("        refresh %d IO files\n", m_TotalIONum);
    printf("        refresh %d NonIO files\n", m_TotalFileNum);
    printf("        refresh %d dirs\n", m_TotalDirNum);
    printf("        refresh %d symlinks\n", m_TotalSymlinkNum);

    /* allocate space and set initial value for IOSize */
    m_IONum = new int [m_SubDirNum];
    m_IOSize = new int * [m_SubDirNum];
    int num = m_TotalIONum / m_SubDirNum;
    int remainder = m_TotalIONum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_IONum[i] = (i<remainder) ? num+1 : num;
        m_IOSize[i] = new int [m_IONum[i]];
    }

    foundIO = new BOOL [num+1];

    for (i=j=k=dIndex=iIndex=0; k<m_TotalIONum; k++)
    {
        m_IOSize[i][j] = m_IODist.item[iIndex].size * KB;
        dIndex ++;
        if (dIndex == m_IODist.item[iIndex].pcnt)
        {
            iIndex ++;
            if (iIndex == m_IODist.num)
                dIndex = iIndex = 0;
        }
        i ++;
        if (i == m_SubDirNum)
        {
            i = 0;
            j ++;
        }
    }

    /* allocate space and set initial value for FileMode */
    m_FileNum = new int [m_SubDirNum];
    m_FileMode = new ushort * [m_SubDirNum];
    num = m_TotalFileNum / m_SubDirNum;
    remainder = m_TotalFileNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_FileNum[i] = (i<remainder) ? num+1 : num;
        m_FileMode[i] = new ushort [m_FileNum[i]];
    }

    foundFile = new BOOL [num+1];

    for (i=0; i<m_SubDirNum; i++)
        for (j=0; j<m_FileNum[i]; j++)
        {
#ifndef WIN32
            m_FileMode[i][j] = (ushort) 00644;
#else
            m_FileMode[i][j] = (ushort) (FILE_ATTRIBUTE_ARCHIVE | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_READONLY);
#endif
        }

    /* allocate space and set initial value for DirMode */
    m_DirNum = new int [m_SubDirNum];
    m_DirMode = new uchar * [m_SubDirNum];
    num = m_TotalDirNum / m_SubDirNum;
    remainder = m_TotalDirNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_DirNum[i] = (i<remainder) ? num+1 : num;
        m_DirMode[i] = new uchar [SIZE_BIT_TO_BYTE(m_DirNum[i])];
    }

    foundDir = new BOOL [num+1];

    for (i=0; i<m_SubDirNum; i++)
    {
        num = SIZE_BIT_TO_BYTE(m_DirNum[i]);
        for (j=0; j<num; j++)
            m_DirMode[i][j] = (uchar) 0;
    }

    /* allocate space and set initial value for SymlinkMode */
    m_SymlinkNum = new int [m_SubDirNum];
    m_SymlinkMode = new uchar * [m_SubDirNum];
    num = m_TotalSymlinkNum / m_SubDirNum;
    remainder = m_TotalSymlinkNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_SymlinkNum[i] = (i<remainder) ? num+1 : num;
        m_SymlinkMode[i] = new uchar [SIZE_BIT_TO_BYTE(m_SymlinkNum[i])];
    }

    foundSymlink = new BOOL [num+1];

    for (i=0; i<m_SubDirNum; i++)
    {
        num = SIZE_BIT_TO_BYTE(m_SymlinkNum[i]);
        for (j=0; j<num; j++)
            m_SymlinkMode[i][j] = (uchar) 0;
    }

    /* search and modify the files, dirs, and symlinks in each sub directory, one by one */
    int last_progress = 5;      /* set the initial value of the job progress */
    int progress;
    for (i=0; i<m_SubDirNum; i++)
    {
        if (((i + 1) % 10) == 0)
            printf("Refresh sub-dir %d of %d\n", i+1, m_SubDirNum);

        /* set all found tag as FALSE */
        for (j=0; j<m_IONum[i]; j++)
            foundIO[j] = FALSE;

        for (j=0; j<m_FileNum[i]; j++)
            foundFile[j] = FALSE;

        for (j=0; j<m_DirNum[i]; j++)
            foundDir[j] = FALSE;

        for (j=0; j<m_SymlinkNum[i]; j++)
            foundSymlink[j] = FALSE;

        /* get this sub directory name */
        GetSubName(dir_name, i);

        /* open this sub directory */
#ifndef WIN32
        DIR *dirp = opendir(dir_name);
        if (dirp == NULL)
#else
        lstrcpy(file_name, dir_name);
        lstrcat(file_name, "\\*.*");
        WIN32_FIND_DATA FileData;
        HANDLE hSearch = FindFirstFile(file_name, &FileData);
        if (hSearch == INVALID_HANDLE_VALUE)
#endif
        {
            fprintf(stderr, "Make the sub direcotry (%d) first\n", i);
            continue;
        }

        /* search this sub directory */
#ifndef WIN32
        struct dirent *dp;
        while ((dp = readdir(dirp)) != NULL)
        {
            strcpy(name, dp->d_name);
#else
        /* serarch this directory */
        do 
        {

            strcpy(name, FileData.cFileName);
#endif
            if (name[0] == '.')     /* ignore "." && ".." */
                continue;
            j = atoi(name+1);
            if (j == 0 && (strlen(name+1) != 6 || !isdigit(name[1])))
                j = -1;
            deleteTag = TRUE;

            if (name[0] == 'F' && (j >= 0 && j < m_IONum[i]))
            {
                deleteTag = FALSE;
                foundIO[j] = TRUE;
                GetIOFileName(name, i, j);
                CSafeFileOperation::Lookup(name, &size, NULL);
                if (size > m_IOSize[i][j])
                    CSafeFileOperation::Truncate(name, m_IOSize[i][j]);
                else if (size < m_IOSize[i][j])
                    CSafeFileOperation::Append(name, m_IOSize[i][j] - size);
            }
            else if (name[0] == 'G' && (j >= 0 && j < m_FileNum[i]))
            {
                deleteTag = FALSE;
                foundFile[j] = TRUE;
                GetNonIOFileName(name, i, j);
                CSafeFileOperation::Chmod(name, m_FileMode[i][j]);
            }
            else if (name[0] == 'D' && (j >= 0 && j < m_DirNum[i]))
            {
                deleteTag = FALSE;
                foundDir[j] = TRUE;
                m_DirMode[i][j/8] |= (1 << (j % 8));
            }
            else if (name[0] == 'L' && (j >= 0 && j < m_SymlinkNum[i]))
            {
                deleteTag = FALSE;
                foundSymlink[j] = TRUE;
                m_SymlinkMode[i][j/8] |= (1 << (j % 8));
            }

            if (deleteTag)
            {
                strcpy(file_name, dir_name);
                strcat(file_name, _STR_PATHSEP);
                strcat(file_name, name);
                CSafeFileOperation::RemoveAny(file_name);
            }
#ifndef WIN32
        }
        closedir(dirp);
#else
        } while (FindNextFile(hSearch, &FileData));
        FindClose(hSearch);
#endif

        /* create new IO files */
        for (j=0; j<m_IONum[i]; j++)
        {
            if (! foundIO[j])
            {
                GetIOFileName(name, i, j);
                CSafeFileOperation::CreateWrite(name, m_IOSize[i][j]);
            }
        }

        /* create new NonIO files */
        for (j=0; j<m_FileNum[i]; j++)
        {
            if (! foundFile[j])
            {
                GetNonIOFileName(name, i, j);
                CSafeFileOperation::CreateNull(name, m_FileMode[i][j]);
            }
        }

        /* create new dirs */
        for (j=0; j<m_DirNum[i]; j++)
        {
            if (! foundDir[j])
            {
                GetDirName(name, i, j);
                CSafeFileOperation::Mkdir(name);
                m_DirMode[i][j/8] |= (1 << (j % 8));
            }
        }

        /* create new NonIO files */
        for (j=0; j<m_SymlinkNum[i]; j++)
        {
            if (! foundSymlink[j])
            {
                GetSymlinkName(name, i, j);
                GetSymlinkTargetName(target_name);
                CSafeFileOperation::Symlink(name, target_name);
                m_SymlinkMode[i][j/8] |= (1 << (j % 8));
            }
        }

        if (m_bUpdate)
        {
            progress = 5 +  (i+1) * 95 / m_SubDirNum;
            if (progress - last_progress >= 5)
            {
                UpdateProgress(progress, FALSE);
                last_progress = progress;
            }
        }
    }

    delete [] foundIO;
    delete [] foundFile;
    delete [] foundDir;
    delete [] foundSymlink;
}


/* create IO files, create NonIO files, create dirs, create NonIO files */
void CClientFileSysBase::Refresh_Content_Multithread()
{
    int i, j, k, dIndex, iIndex;
    int foundIO_Num, foundFile_Num, foundDir_Num, foundSymlink_Num;

    printf("In CClientFileSysBase::Refresh_Content_Multithread()\n");
    printf("        refresh %d IO files\n", m_TotalIONum);
    printf("        refresh %d NonIO files\n", m_TotalFileNum);
    printf("        refresh %d dirs\n", m_TotalDirNum);
    printf("        refresh %d symlinks\n", m_TotalSymlinkNum);

    /* allocate space and set initial value for IOSize */
    m_IONum = new int [m_SubDirNum];
    m_IOSize = new int * [m_SubDirNum];
    int num = m_TotalIONum / m_SubDirNum;
    int remainder = m_TotalIONum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_IONum[i] = (i<remainder) ? num+1 : num;
        m_IOSize[i] = new int [m_IONum[i]];
    }

    foundIO_Num = num + 1;

    for (i=j=k=dIndex=iIndex=0; k<m_TotalIONum; k++)
    {
        m_IOSize[i][j] = m_IODist.item[iIndex].size * KB;
        dIndex ++;
        if (dIndex == m_IODist.item[iIndex].pcnt)
        {
            iIndex ++;
            if (iIndex == m_IODist.num)
                dIndex = iIndex = 0;
        }
        i ++;
        if (i == m_SubDirNum)
        {
            i = 0;
            j ++;
        }
    }

    /* allocate space and set initial value for FileMode */
    m_FileNum = new int [m_SubDirNum];
    m_FileMode = new ushort * [m_SubDirNum];
    num = m_TotalFileNum / m_SubDirNum;
    remainder = m_TotalFileNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_FileNum[i] = (i<remainder) ? num+1 : num;
        m_FileMode[i] = new ushort [m_FileNum[i]];
    }

    foundFile_Num = num + 1;

    for (i=0; i<m_SubDirNum; i++)
        for (j=0; j<m_FileNum[i]; j++)
        {
#ifndef WIN32
            m_FileMode[i][j] = (ushort) 00644;
#else
            m_FileMode[i][j] = (ushort) (FILE_ATTRIBUTE_ARCHIVE | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_READONLY);
#endif
        }

    /* allocate space and set initial value for DirMode */
    m_DirNum = new int [m_SubDirNum];
    m_DirMode = new uchar * [m_SubDirNum];
    num = m_TotalDirNum / m_SubDirNum;
    remainder = m_TotalDirNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_DirNum[i] = (i<remainder) ? num+1 : num;
        m_DirMode[i] = new uchar [SIZE_BIT_TO_BYTE(m_DirNum[i])];
    }

    foundDir_Num = num + 1;

    for (i=0; i<m_SubDirNum; i++)
    {
        num = SIZE_BIT_TO_BYTE(m_DirNum[i]);
        for (j=0; j<num; j++)
            m_DirMode[i][j] = (uchar) 0;
    }

    /* allocate space and set initial value for SymlinkMode */
    m_SymlinkNum = new int [m_SubDirNum];
    m_SymlinkMode = new uchar * [m_SubDirNum];
    num = m_TotalSymlinkNum / m_SubDirNum;
    remainder = m_TotalSymlinkNum % m_SubDirNum;
    for (i=0; i<m_SubDirNum; i++)
    {
        m_SymlinkNum[i] = (i<remainder) ? num+1 : num;
        m_SymlinkMode[i] = new uchar [SIZE_BIT_TO_BYTE(m_SymlinkNum[i])];
    }

    foundSymlink_Num = num + 1;

    for (i=0; i<m_SubDirNum; i++)
    {
        num = SIZE_BIT_TO_BYTE(m_SymlinkNum[i]);
        for (j=0; j<num; j++)
            m_SymlinkMode[i][j] = (uchar) 0;
    }

    /* create working threads and put them into thread pool */
    CClientFileSysRefPool pool;
#ifndef WIN32
    int thread_num = 24;
#else
    int thread_num = 16;
#endif
    CClientFileSysRefThread **threads = new CClientFileSysRefThread * [thread_num];
    for (i=0; i<thread_num; i++)
    {
        threads[i] = new CClientFileSysRefThread(i, &pool, this);
        threads[i]->SetFSBaseData(m_IONum, m_IOSize, m_FileNum, m_FileMode, m_DirNum, m_DirMode, m_SymlinkNum, m_SymlinkMode);
        threads[i]->SetFoundNum(foundIO_Num, foundFile_Num, foundDir_Num, foundSymlink_Num);
        threads[i]->Start();
    }
    pool.AddFileSysRefThread(thread_num, threads);
    delete [] threads;

    /* search and modify the files, dirs, and symlinks in each sub directory, one by one */
    int last_progress = 5;      /* set the initial value of the job progress */
    int progress;
    CClientFileSysRefThread *idle_thread;
    for (i=0; i<m_SubDirNum; i++)
    {
        if (((i + 1) % 10) == 0)
            printf("Refresh sub-dir %d of %d\n", i+1, m_SubDirNum);

        idle_thread = pool.GetIdleFileSysRefThread();
        idle_thread->SetSubdirAndStart(i);
//printf("Use thread %02d to refresh sub-dir %d\n", idle_thread->m_id, i);

        if (m_bUpdate)
        {
            progress = 5 +  (i+1) * 95 / m_SubDirNum;
            if (progress - last_progress >= 5)
            {
                UpdateProgress(progress, FALSE);
                last_progress = progress;
            }
        }

        // check break
        if (IsBreak())
            break;
    }

    pool.WairForAllIdle();
    for (i=0; i<thread_num; i++)
        pool.GetFileSysRefThreadByID(i)->Quit();
    _SLEEP(2);  /* wait for all thread terminate */
}

BOOL CClientFileSysBase::UpdateProgress(int progress, BOOL done)
{
    CSocketClient socket;
    /* first connect to resource manager server */
    if (! socket.ConnectToServer(m_HostIp, REQ_SERVER_PORT))
    {
        fprintf(stderr, "Cannot connect to resource manager server (%s).\n", m_HostName);
        return FALSE;
    }

    CPacket packet;
    char *buffer;
    int size;

    /* message type */
    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_FAUCET_REFSERVERUPDATE;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* updated progress information */
    CNet_ServerRefUpdate net_update;
    net_update.job_id = m_job_id;
    net_update.client_id = m_client_id;
    net_update.progress = progress;
    net_update.done = done;
    net_update.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* if this job is done, need to send the file system information */
    if (done)
    {
        /* send back the file system information to RM */
        printf("Sending the file system information to resource manager server ...\n");
        SendFileSysInfoToHost(&socket);
        printf("Finished ...\n");
    }

    return TRUE;
}


int CClientFileSysBase::RoundSubDirNum(int num)
{
    static int base[] = {
        3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97,
        101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193,
        197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307,
        311, 313, 317, 331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419, 421,
        431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509, 521, 523, 541, 547,
        557, 563, 569, 571, 577, 587, 593, 599, 601, 607, 613, 617, 619, 631, 641, 643, 647, 653, 659,
        661, 673, 677, 683, 691, 701, 709, 719, 727, 733, 739, 743, 751, 757, 761, 769, 773, 787, 797,
        809, 811, 821, 823, 827, 829, 839, 853, 857, 859, 863, 877, 881, 883, 887, 907, 911, 919, 929,
        937, 941, 947, 953, 967, 971, 977, 983, 991, 997
    };

    int max = 997;
    int i = 0;
    while (base[i] < max)
    {
        if (base[i] >= num)
            return base[i];
        i ++;
    }
    return max;
}


//DEBUG
//#define DEBUG_SMALL_FS


void CClientFileSysBase::CalFileNum()
{
#ifndef WIN32
    struct statvfs vfs;

    if (statvfs(m_LocalPath, &vfs) == -1)
        return;

    if (vfs.f_frsize < KB)
        m_TotalSize = vfs.f_blocks / (KB / vfs.f_frsize);
    else
        m_TotalSize = vfs.f_blocks * (vfs.f_frsize / KB);
    m_TotalFNodeNum = vfs.f_files;
#else
    BOOL ret;
    char RootPath[MAX_PATHLEN], *chptr;
    DWORD SectorsPerCluster;
    DWORD BytesPerSector;
    DWORD NumberOfFreeClusters;
    DWORD TotalNumberOfClusters;
    DWORD TotalSize;

    strcpy(RootPath, m_LocalPath);
    ret = TRUE;
    chptr = strchr(RootPath, ':');
    if (chptr != NULL)
    {
        *(chptr+1) = '\\';
        *(chptr+2) = 0;
    }
    else if (strncmp(RootPath, "\\\\", 2) == 0)
    {
        chptr = strchr(RootPath+2, '\\');
        if (chptr != NULL)
        {
            chptr = strchr(chptr+1, '\\');
            if (chptr != NULL)
                *(chptr+1) = 0;
            else
                strcat(RootPath, "\\");
        }
        else
            ret = FALSE;
    }
    else
        ret = FALSE;

    if (! ret)
        return;


    if (! GetDiskFreeSpace(RootPath,            // root path
                        &SectorsPerCluster,     // sectors per cluster
                        &BytesPerSector,        // bytes per sector
                        &NumberOfFreeClusters,  // free clusters
                        &TotalNumberOfClusters))// total clusters
        return;

    if (BytesPerSector < KB)
        TotalSize = TotalNumberOfClusters * SectorsPerCluster / (KB / BytesPerSector);
    else
        TotalSize = TotalNumberOfClusters * SectorsPerCluster * (BytesPerSector / KB);
    m_TotalSize = (int) TotalSize;

    /* use file system size to estimate inode number (inode:size(KB) <= 0.12) */
    m_TotalFNodeNum = (int) (TotalSize * 0.12);
#endif

#ifdef DEBUG_SMALL_FS
    m_TotalSize = MB;
    m_TotalFNodeNum = 200000;
#endif

    double avg_size = 0.0;      /* get one average file size, in KBytes */
    int prev_pcnt = 0;
    for (int i=0; i<m_IODist.num; i++)
    {
        avg_size += (double) ((m_IODist.item[i].pcnt - prev_pcnt) * m_IODist.item[i].size);
        prev_pcnt = m_IODist.item[i].pcnt;
    }
    avg_size *= 0.01;

    double UsedSize = (double) m_TotalSize * 0.4;                   /* only 40% space used for IO files */
    double LargestSize;
    if (avg_size < 100.0)
        LargestSize = 2 * 1024 * 1024;      /* small file set, 2G */
    else if (avg_size < 1000.0)
        LargestSize = 3 * 1024 * 1024;      /* medium file set, 3G */
    else
        LargestSize = 5 * 1024 * 1024;      /* large file set, 5G */
    if (UsedSize > LargestSize)
        UsedSize = LargestSize;
    m_TotalIONum = (int) (UsedSize / avg_size);

    m_SubDirNum = m_TotalIONum / 10;                                /* at least 10 IO files in each directory */
    if (m_SubDirNum > SUBDIR_MAXMNUM)
        m_SubDirNum = SUBDIR_MAXMNUM;
    else
        m_SubDirNum = RoundSubDirNum(m_SubDirNum);

    m_TotalFNodeNum -= m_TotalIONum;

    m_TotalFileNum = m_TotalFNodeNum / 5;                    /* 20% file node used for NonIO files */
    if (m_TotalFileNum > MAX_FILEINDEX / 10)                 /* don't exceed 10% of max file index */
        m_TotalFileNum = MAX_FILEINDEX / 10;

    m_TotalDirNum = m_TotalFNodeNum / 5;                     /* 20% file node used for directory */
    if (m_TotalDirNum > MAX_DIRINDEX / 10)                   /* don't exceed 10% of max dir index */
        m_TotalDirNum = MAX_DIRINDEX / 10;

#ifndef WIN32
    m_TotalSymlinkNum = m_TotalFNodeNum / 10;                /* 10% file node used for directory */
    if (m_TotalSymlinkNum > MAX_SYMLINKINDEX / 10)           /* don't exceed 10% of max symlink index */
        m_TotalSymlinkNum = MAX_SYMLINKINDEX / 10;
#else
    m_TotalSymlinkNum = 0;                                   /* no symlink on windows platform */
#endif


#ifdef DEBUG_SMALL_FS

    m_SubDirNum = 47;
    m_TotalIONum = 10000;
    m_TotalFileNum = 4000;
    m_TotalDirNum = 4000;
//    m_SubDirNum = 10;
//    m_TotalIONum = 100;
//    m_TotalFileNum = 40;
//    m_TotalDirNum = 40;
#ifndef WIN32
    m_TotalSymlinkNum = 4000;
#else
    m_TotalSymlinkNum = 0;
#endif

    /*m_SubDirNum = 997;
    m_TotalIONum = 60000;
    m_TotalFileNum = 50000;
    m_TotalDirNum = 50000;
#ifndef WIN32
    m_TotalSymlinkNum = 50000;
#else
    m_TotalSymlinkNum = 0;
#endif*/

#endif

    printf("In CClientFileSysBase::CalFileNum()\n");
    printf("    TotalSize     = %d KB\n", m_TotalSize);
    printf("    TotalFNodeNum = %d\n", m_TotalFNodeNum);
    printf("    SubDirNum     = %d\n", m_SubDirNum);
    printf("    IONum         = %d\n", m_TotalIONum);
    printf("    FileNum       = %d\n", m_TotalFileNum);
    printf("    DirNum        = %d\n", m_TotalDirNum);
    printf("    SymlinkNum    = %d\n", m_TotalSymlinkNum);
}

BOOL CClientFileSysBase::SendFileSysInfoToHost(CSocketClient *socket)
{
    /* fill into network exchange data structure */
    CNet_FsInfo net_fsinfo;
    net_fsinfo.TotalSize = m_TotalSize;
    net_fsinfo.TotalFNodeNum = m_TotalFNodeNum;
//    net_fsinfo.IODist = m_IODist;
    net_fsinfo.SubDirNum = m_SubDirNum;
    net_fsinfo.TotalIONum = m_TotalIONum;
    net_fsinfo.TotalFileNum = m_TotalFileNum;
    net_fsinfo.TotalDirNum = m_TotalDirNum;
    net_fsinfo.TotalSymlinkNum = m_TotalSymlinkNum;
    net_fsinfo.IONum = m_IONum;
    net_fsinfo.IOSize = m_IOSize;
    net_fsinfo.FileNum = m_FileNum;
    net_fsinfo.FileMode = m_FileMode;
    net_fsinfo.DirNum = m_DirNum;
    net_fsinfo.DirMode = m_DirMode;
    net_fsinfo.SymlinkNum = m_SymlinkNum;
    net_fsinfo.SymlinkMode = m_SymlinkMode;

    /* put into data buffer */
    char *buffer;
    int size;
    net_fsinfo.xdt_trans(XDTOP_PUT, buffer, size);

    /* packet and send the data */
    CPacket packet(buffer, size);
    if (! packet.Send(socket))
        return FALSE;

    return TRUE;
}

BOOL CClientFileSysBase::SendFileSysInfoToPrim()
{
    CSocketClient socket;
    /* first connect to resource manager server */
    if (! socket.ConnectToServer(m_HostIp, CLIENT_MAIN_PORT))
    {
        fprintf(stderr, "Cannot connect to prim client (%s).\n", m_HostName);
        return FALSE;
    }

    CPacket packet;
    char *buffer;
    int size;

    /* send the message type first */
    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_PUMPER_REF_FSINFO;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* then send the load id, use job_id as load_id */
    CNet_LoadID net_loadid;
    net_loadid.id = m_job_id;
    net_loadid.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* then send the client id */
    CNet_ClientID net_clientid;
    net_clientid.id = m_client_id;
    net_clientid.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    return SendFileSysInfoToHost(&socket);
}
