
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysInfo.cpp
**
** DESCRIPTION: This class mainly manage the file system information used
**              by the client.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientFileSysInfo::CClientFileSysInfo()
**     CClientFileSysInfo::~CClientFileSysInfo()
**
**
****************************************************************************/


#ifndef WIN32
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include <math.h>
#include <limits.h>
#include "Common.h"
#include "ClientFileSysInfo.h"
#include "SafeFileOperation.h"
#include "SysTime.h"

CClientFileSysInfo::CClientFileSysInfo(CClientFileSysBase *fsbase)
{
    m_fsbase = fsbase;
    m_IOSizeLevel = 1;

    InitMutex(m_IO_lock);
    InitMutex(m_NonIO_lock);
    InitMutex(m_Dir_lock);
    InitMutex(m_Symlink_lock);

    m_IOSize = NULL;
    m_IOAcceSet = NULL;
    m_IOAcceRange = NULL;
    m_FileMode = NULL;
    m_DirMode = NULL;
    m_SymlinkMode = NULL;
}

CClientFileSysInfo::~CClientFileSysInfo()
{
#ifdef WIN32
    if (m_IO_lock != NULL)
        CloseHandle(m_IO_lock);
    if (m_NonIO_lock != NULL)
        CloseHandle(m_NonIO_lock);
    if (m_Dir_lock != NULL)
        CloseHandle(m_Dir_lock);
    if (m_Symlink_lock != NULL)
        CloseHandle(m_Symlink_lock);
#endif

    if (m_IOSize != NULL)
        delete [] m_IOSize;
    if (m_IOAcceSet != NULL)
        delete [] m_IOAcceSet;
    if (m_IOAcceRange != NULL)
        delete [] m_IOAcceRange;
    if (m_FileMode != NULL)
        delete [] m_FileMode;
    if (m_DirMode != NULL)
        delete [] m_DirMode;
    if (m_SymlinkMode != NULL)
        delete [] m_SymlinkMode;
}

void CClientFileSysInfo::InitSubDir(int num, int *subdir)
{
    m_OptDirNum = num;
    for (int i=0; i<num; i++)
        m_OptDir[i] = subdir[i];
}

void CClientFileSysInfo::InitIOFileInfo(int *num, int **size)
{
    int i, j, k, max_i, max_num;

    max_num = -1;
    for (i=0; i<m_OptDirNum; i++)
    {
        m_IONum[i] = num[i];
        if (num[i] >= max_num)
        {
            max_num = num[i];
            max_i = i;
        }
    }

    if (m_IOSize != NULL)
        delete [] m_IOSize;
    m_IOSize = new int [m_OptDirNum * max_num];

    k = 0;
    for (j=0; j<max_num; j++)
    {
        for (i=0; i<m_OptDirNum; i++)
        {
            if (j >= num[i])
                m_IOSize[k ++] = 0;
            else
                m_IOSize[k ++] = size[i][j];
        }
    }
}

void CClientFileSysInfo::InitNonIOFileInfo(int *num, ushort **mode)
{
    int i, j, k, n, max_i, max_num;

    max_num = -1;
    for (i=0; i<m_OptDirNum; i++)
    {
        /* get actual number for each directory */
        for (j=num[i]-1; j>=0; j--)
        {
            if ((mode[i][j] & 0x200) == 0)
                break;
        }
        /* actual number is j+1 */
        if (j + 1 >= max_num)
        {
            max_num = j + 1;
            max_i = i;
        }
    }

    m_NewFile_Index = (max_num-1) * m_OptDirNum + max_i + 1;
    m_MaxFileNum = MAX_FILEINDEX;
    if (m_FileMode != NULL)
        delete [] m_FileMode;
    m_FileMode = new ushort [m_MaxFileNum];

    m_TotalFileNum = 0;
    k = 0;
    for (j=0; j<max_num; j++)
    {
        n = (j==max_num-1) ? max_i+1 : m_OptDirNum;
        for (i=0; i<n; i++)
        {
            if (j >= num[i] || (mode[i][j] & 0x0200) != 0)
                m_FileMode[k ++] = (ushort) 0x0200;
            else
            {
                m_FileMode[k ++] = (ushort) (mode[i][j] & 0x01FF);
                m_TotalFileNum ++;
            }
        }
    }

    m_MinFileNum = m_TotalFileNum / 20;     /* need at least 5% file exist */
    if (m_MinFileNum < 10)
        m_MinFileNum = 10;
}

void CClientFileSysInfo::InitDirInfo(int *num, uchar **mode)
{
    int i, j, k, n, max_i, max_num;

    max_num = -1;
    for (i=0; i<m_OptDirNum; i++)
    {
        /* get actual number for each directory */
        for (j=num[i]-1; j>=0; j--)
        {
            if (((mode[i][j/8] >> (j % 8)) & 0x01) != 0)
                break;
        }
        /* actual number is j+1 */
        if (j + 1 >= max_num)
        {
            max_num = j + 1;
            max_i = i;
        }
    }

    m_NewDir_Index = (max_num-1) * m_OptDirNum + max_i + 1;
    m_MaxDirNum = MAX_DIRINDEX;
    if (m_DirMode != NULL)
        delete [] m_DirMode;
    m_DirMode = new uchar [m_MaxDirNum / 4];  /* 2 bits for each dir, 1 byte for 4 dirs */

    m_TotalDirNum = 0;
    k = 0;
    for (j=0; j<max_num; j++)
    {
        n = (j==max_num-1) ? max_i+1 : m_OptDirNum;
        for (i=0; i<n; i++)
        {
            if ((k % 4) == 0)
                m_DirMode[k/4] = 0;
            if (j < num[i] && ((mode[i][j/8] >> (j % 8)) & 0x01) != 0)
            {
                m_DirMode[k/4] |= 1 << ((k % 4) * 2);  /* dir exist */
                m_TotalDirNum ++;
            }
            k ++;
        }
    }

    m_MinDirNum = m_TotalDirNum / 20;     /* need at least 5% dir exist */
    if (m_MinDirNum < 10)
        m_MinDirNum = 10;
}

void CClientFileSysInfo::InitSymlinkInfo(int *num, uchar **mode)
{
    int i, j, k, n, max_i, max_num;

    max_num = -1;
    for (i=0; i<m_OptDirNum; i++)
    {
        /* get actual number for each symlink */
        for (j=num[i]-1; j>=0; j--)
        {
            if (((mode[i][j/8] >> (j % 8)) & 0x01) != 0)
                break;
        }
        /* actual number is j+1 */
        if (j + 1 >= max_num)
        {
            max_num = j + 1;
            max_i = i;
        }
    }

    m_NewSymlink_Index = (max_num-1) * m_OptDirNum + max_i + 1;
    m_MaxSymlinkNum = MAX_SYMLINKINDEX;
    if (m_SymlinkMode != NULL)
        delete [] m_SymlinkMode;
    m_SymlinkMode = new uchar [m_MaxSymlinkNum / 4];  /* 2 bits for each symlink, 1 byte for 4 symlinks */

    m_TotalSymlinkNum = 0;
    k = 0;
    for (j=0; j<max_num; j++)
    {
        n = (j==max_num-1) ? max_i+1 : m_OptDirNum;
        for (i=0; i<n; i++)
        {
            if ((k % 4) == 0)
                m_SymlinkMode[k/4] = 0;
            if (j < num[i] && ((mode[i][j/8] >> (j % 8)) & 0x01) != 0)
            {
                m_SymlinkMode[k/4] |= 1 << ((k % 4) * 2);  /* symlink exist */
                m_TotalSymlinkNum ++;
            }
            k ++;
        }
    }

    m_MinSymlinkNum = m_TotalSymlinkNum / 20;     /* need at least 5% symlink exist */
    if (m_MinSymlinkNum < 10)
        m_MinSymlinkNum = 10;
}

void CClientFileSysInfo::InitIOFileAcce()
{
    int i, j, k;

    int totalnum =0;
    for (i=0; i<m_OptDirNum; i++)
        totalnum += m_IONum[i];

    m_IOAcceNum = (int) (totalnum * IOACCESS_PERCENT);

    m_IOAcceSet = new int [m_IOAcceNum];

    int *indexOrder = new int [totalnum];
    k = 0;
    for (i=0; i<m_OptDirNum; i++)
    {
        for (j=0; j<m_IONum[i]; j++)
            indexOrder[k++] = j*m_OptDirNum + i;
    }
    ShuffleArray(totalnum, indexOrder);
    for (i=0; i<m_IOAcceNum; i++)
        m_IOAcceSet[i] = indexOrder[i];
    delete [] indexOrder;

    m_IOAcceGroupNum = (m_IOAcceNum < 8) ? m_IOAcceNum : 8 + (m_IOAcceNum/500) * 4;
    m_IOAccesGroupSize = m_IOAcceNum / m_IOAcceGroupNum;
    m_IOAcceRange = new int [m_IOAcceGroupNum];

    int range = 0;
    double lambda = (double) (m_IOAcceGroupNum / 2);
    if (lambda <= 0.0)
        lambda = 1.0;
    double e_lambda = exp(lambda);
    double cumulative_lambda_power = 1.0;
    double cumulative_x_factorial = 1.0;
    int next_value;

    /* assign a range to each group */
    for (i=0; i<m_IOAcceGroupNum; i++)
    {
        /* get next value in poisson distribution sequence, using */
        /*    lambda^x / (e^(lambda) * x!) , for x=1,2,3,...,m_IOAcceGroupNum */
        next_value = (int) (1000 * (cumulative_lambda_power / (e_lambda * cumulative_x_factorial)));
        range += next_value;
        m_IOAcceRange[i] = range;

        cumulative_lambda_power *= (double) lambda;
        cumulative_x_factorial *= (double) (i+1);
    }
    m_IOAcceMaxRange = range;

    m_IsBestFit = FALSE;
}

void CClientFileSysInfo::CollSubDir(int &num, int *&subdir)
{
    num = m_OptDirNum;
    subdir = new int [m_OptDirNum];
    memcpy(subdir, m_OptDir, m_OptDirNum*sizeof(int));
}

void CClientFileSysInfo::CollIOFileInfo(int *&num, int **&size)
{
    num = new int [m_OptDirNum];
    memcpy(num, m_IONum, m_OptDirNum*sizeof(int));
    size = new int * [m_OptDirNum];
    int i, j, k;
    for (i=0; i<m_OptDirNum; i++)
    {
        size[i] = new int [m_IONum[i]];
        k = i;
        for (j=0; j<m_IONum[i]; j++)
        {
            size[i][j] = m_IOSize[k];
            k += m_OptDirNum;
        }
    }
}

void CClientFileSysInfo::CollNonIOFileInfo(int &totalnum, int *&num, ushort **&mode)
{
    totalnum = m_TotalFileNum;
    num = new int [m_OptDirNum];
    int basenum = m_NewFile_Index / m_OptDirNum;
    int remnum = m_NewFile_Index % m_OptDirNum;
    int i, j, k;
    for (i=0; i<remnum; i++)
        num[i] = basenum + 1;
    for (i=remnum; i<m_OptDirNum; i++)
        num[i] = basenum;
    mode = new ushort * [m_OptDirNum];
    for (i=0; i<m_OptDirNum; i++)
    {
        mode[i] = new ushort [num[i]];
        k = i;
        for (j=0; j<num[i]; j++)
        {
            mode[i][j] = m_FileMode[k];
            k += m_OptDirNum;
        }
    }
}

void CClientFileSysInfo::CollDirInfo(int &totalnum, int *&num, uchar **&mode)
{
    totalnum = m_TotalDirNum;
    num = new int [m_OptDirNum];
    int basenum = m_NewDir_Index / m_OptDirNum;
    int remnum = m_NewDir_Index % m_OptDirNum;
    int i, j, k;
    for (i=0; i<remnum; i++)
        num[i] = basenum + 1;
    for (i=remnum; i<m_OptDirNum; i++)
        num[i] = basenum;
    mode = new uchar * [m_OptDirNum];
    for (i=0; i<m_OptDirNum; i++)
    {
        mode[i] = new uchar [SIZE_BIT_TO_BYTE(num[i])];
        k = i;
        for (j=0; j<num[i]; j++)
        {
            if ((j % 8) == 0)
                mode[i][j/8] = (uchar) 0x00;
            mode[i][j/8] |= ((m_DirMode[k/4] >> (2 * (k % 4))) & 0x01) << (j % 8);
            k += m_OptDirNum;
        }
    }
}

void CClientFileSysInfo::CollSymlinkInfo(int &totalnum, int *&num, uchar **&mode)
{
    totalnum = m_TotalSymlinkNum;
    num = new int [m_OptDirNum];
    int basenum = m_NewSymlink_Index / m_OptDirNum;
    int remnum = m_NewSymlink_Index % m_OptDirNum;
    int i, j, k;
    for (i=0; i<remnum; i++)
        num[i] = basenum + 1;
    for (i=remnum; i<m_OptDirNum; i++)
        num[i] = basenum;
    mode = new uchar * [m_OptDirNum];
    for (i=0; i<m_OptDirNum; i++)
    {
        mode[i] = new uchar [SIZE_BIT_TO_BYTE(num[i])];
        k = i;
        for (j=0; j<num[i]; j++)
        {
            if ((j % 8) == 0)
                mode[i][j/8] = (uchar) 0x00;
            mode[i][j/8] |= ((m_SymlinkMode[k/4] >> (2 * (k % 4))) & 0x01) << (j % 8);
            k += m_OptDirNum;
        }
    }
}

BOOL CClientFileSysInfo::GetIOFileInfo(int req_size, char *name, int &handle, int &size)
{
    int value = GetRandom() % m_IOAcceMaxRange;

    int low = 0;
    int high = m_IOAcceGroupNum - 1;
    int previous = -1;
    int group;
    while ( 1 )
    {
        group = (low + high) / 2;
        if (previous == group)
            break;
        if (m_IOAcceRange[group] == value)
            break;
        if (m_IOAcceRange[group] > value)
        {
            if (group == 0)
                break;
            if (m_IOAcceRange[group-1] < value)
                break;
            else
                high = group - 1;
        }
        else /* m_IOAcceRange[group] < value */
        {
            if (m_IOAcceRange[group+1] > value)
            {
                group ++;
                break;
            }
            else
                low = group + 1;
        }
        previous = group;
    }

    int start_group, group_size;
    int start_file, file, found_file;
    int index, offset;
    int delta, best_delta;

    int indexName = -1;
    best_delta = INT_MAX;
    start_group = group;
    found_file = 0;

    /* need to lock because multithread may change the m_IOSize to indicate that one file is being used by IO operations */
    LockMutex(m_IO_lock);

    while ( 1 )
    {
        /* Pick a file within the group to operate on. Since  (m_IOAcceNum / m_IOAccesGroupSize)
           may have a remainder, groups may have either m_IOAccesGroupSize or (m_IOAccesGroupSize+1) files. */
        group_size = m_IOAccesGroupSize;
        if (group < (m_IOAcceNum % m_IOAccesGroupSize))
            group_size ++;

        /* Beginning with a random starting point in the group, search for a file that is 
           eligible for this operation. index is an index into the set of all files.
           file and start_file are indices into m_IOAcceSet array. */
        offset = GetRandom() % group_size;
        start_file = group + (offset * m_IOAcceGroupNum);
        file = start_file;
        while ( 1 )
        {
            index = m_IOAcceSet[file];
            size = m_IOSize[index];

            /* if the highest bit is set, the file is being used by a IO operation */
            if ((size & 0x80000000) == 0 && size >= req_size)
            {
                if (! m_IsBestFit || req_size <= 0)
                {
                    indexName = index;
                    break;
                }
                else
                {
                    /*
                    * If we find an good enough match we should use it.
                    * Define good enough to be req_size <= X < req_size + 8K
                    * If not we continue to search for the best fit within
                    * a fixed distance 8.
                    */
                    if (size < (req_size + 8 * 1024))
                    {
                        indexName = index;
                        break;
                    }

                    found_file ++;
                    delta = size - req_size;
                    if (indexName == -1 || delta < best_delta)
                    {
                        indexName = index;
                        best_delta = delta;
                    }

                    if (found_file >= 8)
                        break;
                }
            }
            offset = (offset + 1) % group_size;
            file = group + (offset * m_IOAcceGroupNum);
            if (file == start_file)
                break;
        }

        if (indexName != -1)
            break;
        group = (group + 1) % m_IOAcceGroupNum;
        if (group == start_group)
            break;
    }

    if (indexName != -1)
    {
        handle = indexName;
        int dir_Index = indexName % m_OptDirNum;
        int file_Index = indexName / m_OptDirNum;
        m_fsbase->GetIOFileName(name, m_OptDir[dir_Index], file_Index);
        size = m_IOSize[indexName];

        /* change this file's size (set the highest bit) to indicate that it's being used by an IO operation */
        m_IOSize[indexName] |= 0x80000000;
    }

    /* unlock it */
    UnlockMutex(m_IO_lock);

    return indexName != -1;
}

/* option: 0(access) 1(remove) 2(create) */
void CClientFileSysInfo::SetIOFileInfo(int handle, int size)
{
    /* reset file size, need to lock it because some thread may use this file size to pick a file for IO operations */
    LockMutex(m_IO_lock);
    m_IOSize[handle] = size;
    UnlockMutex(m_IO_lock);
}

/* option: 0(access) 1(remove) 2(create) */
BOOL CClientFileSysInfo::GetNonIOFileInfo(int option, char *name, int &handle, int &mode)
{
    if ((option < 2 && m_TotalFileNum <= m_MinFileNum) || (option == 2 && m_NewFile_Index >= m_MaxFileNum))
        return FALSE;

    int index;

    switch (option)
    {
        case 0:
        case 1:
        {
            LockMutex(m_NonIO_lock);
            while ( 1 )
            {
                index = GetRandom() % m_NewFile_Index;
                if ((m_FileMode[index] & 0x8200) == 0)  /* file deleted or being used */
                    break;
            }
            mode = m_FileMode[index];
            m_FileMode[index] |= 0x8000;                /* set file being used */
            if (option == 1)
            {
                m_FileMode[index] |= 0x0200;            /* set file deleted */
                m_TotalFileNum --;
            }
            UnlockMutex(m_NonIO_lock);

            handle = index;
            break;
        }

        case 2:
            LockMutex(m_NonIO_lock);
            /* change this file's mode (set the highest bit) to indicate that it's being used by an NonIO operation */
            handle = m_NewFile_Index;
            m_FileMode[handle] = 0x8000;
            m_NewFile_Index ++;
            UnlockMutex(m_NonIO_lock);
            mode = 0666;
            m_TotalFileNum ++;
            break;
    }
    int dir_Index = handle % m_OptDirNum;
    int file_Index = handle / m_OptDirNum;
    m_fsbase->GetNonIOFileName(name, m_OptDir[dir_Index], file_Index);

    return TRUE;
}

void CClientFileSysInfo::SetNonIOFileInfo(int option, int handle, int mode)
{
    /* reset file mode, need to lock it because some thread may use this file mode to pick a file for NonIO operations */
    LockMutex(m_NonIO_lock);
    m_FileMode[handle] = mode;
    if (option == 1)
        m_FileMode[handle] |= 0x200;
    UnlockMutex(m_NonIO_lock);
}

/* option: 0(access) 1(remove) 2(create) */
BOOL CClientFileSysInfo::GetDirInfo(int option, char *name, int &handle)
{
    if ((option < 2 && m_TotalDirNum <= m_MinDirNum) || (option == 2 && m_NewDir_Index >= m_MaxDirNum))
        return FALSE;

    int index;

    switch (option)
    {
        case 0:
        case 1:     /* remove */
            LockMutex(m_Dir_lock);
            while ( 1 )
            {
                index = GetRandom() % m_NewDir_Index;
                if (((m_DirMode[index/4] >> ((index%4) * 2)) & 0x03) == 0x01)
                    break;      /* dir exists and not being used */
            }
            m_DirMode[index/4] |= 0x02 << ((index % 4) * 2);        /* set dir being used */
            if (option == 1)
            {
                m_DirMode[index/4] &= ~(0x01 << ((index % 4) * 2)); /* set dir deleted */
                m_TotalDirNum --;
            }
            UnlockMutex(m_Dir_lock);

            handle = index;
            break;

        case 2:     /* create */
            LockMutex(m_Dir_lock);
            /* change this dir's mode (set the two bits as 11) to indicate that it's being used by an IO operation */
            handle = m_NewDir_Index;
            if ((handle % 4) == 0)
                m_DirMode[handle/4] = 0;
            m_DirMode[handle/4] |= 0x03 << ((handle % 4) * 2);  /* directory exist and being used */
            m_NewDir_Index ++;
            m_TotalDirNum ++;
            UnlockMutex(m_Dir_lock);
            break;
    }
    int dir_Index = handle % m_OptDirNum;
    int file_Index = handle / m_OptDirNum;
    m_fsbase->GetDirName(name, m_OptDir[dir_Index], file_Index);

    return TRUE;
}

/* option: 0(access) 1(remove) 2(create) */
void CClientFileSysInfo::SetDirInfo(int option, int handle)
{
    /* for option 1(remove), set two bits as 00; for option 0(access) & 2(create), set two bits as 01 */
    LockMutex(m_Dir_lock);
    m_DirMode[handle/4] &= ~(0x03 << ((handle % 4) * 2));  /* set two bits as 00 */
    if (option != 1)
        m_DirMode[handle/4] |= 0x01 << ((handle % 4) * 2); /* set two bits as 01 */
    UnlockMutex(m_Dir_lock);
}

/* option: 0(access) 1(remove) 2(create) */
BOOL CClientFileSysInfo::GetSymlinkInfo(int option, char *name, int &handle)
{
    if ((option < 2 && m_TotalSymlinkNum <= m_MinSymlinkNum) || (option == 2 && m_NewSymlink_Index >= m_MaxSymlinkNum))
        return FALSE;

    int index;

    switch (option)
    {
        case 0:
        case 1:     /* remove */
            LockMutex(m_Symlink_lock);
            while ( 1 )
            {
                index = GetRandom() % m_NewSymlink_Index;
                if (((m_SymlinkMode[index/4] >> ((index%4) * 2)) & 0x03) == 0x01)
                    break;      /* symlink exists and not being used */
            }
            m_SymlinkMode[index/4] |= 0x02 << ((index % 4) * 2);        /* set symlink being used */
            if (option == 1)
            {
                m_SymlinkMode[index/4] &= ~(0x01 << ((index % 4) * 2)); /* set symlink deleted */
                m_TotalSymlinkNum --;
            }
            UnlockMutex(m_Symlink_lock);

            handle = index;
            break;

        case 2:     /* create */
            LockMutex(m_Symlink_lock);
            /* change this symlink's mode (set the two bits as 11) to indicate that it's being used by an IO operation */
            handle = m_NewSymlink_Index;
            if ((handle % 4) == 0)
                m_SymlinkMode[handle/4] = 0;
            m_SymlinkMode[handle/4] |= 0x03 << ((handle % 4) * 2);  /* symlink exist and being used */
            m_NewSymlink_Index ++;
            m_TotalSymlinkNum ++;
            UnlockMutex(m_Symlink_lock);
            break;
    }
    int dir_Index = handle % m_OptDirNum;
    int file_Index = handle / m_OptDirNum;
    m_fsbase->GetSymlinkName(name, m_OptDir[dir_Index], file_Index);

    return TRUE;
}

/* option: 0(access) 1(remove) 2(create) */
void CClientFileSysInfo::SetSymlinkInfo(int option, int handle)
{
    /* for option 1(remove), set two bits as 00; for option 0(access) & 2(create), set two bits as 01 */
    LockMutex(m_Symlink_lock);
    m_SymlinkMode[handle/4] &= ~(0x03 << ((handle % 4) * 2));  /* set two bits as 00 */
    if (option != 1)
        m_SymlinkMode[handle/4] |= 0x01 << ((handle % 4) * 2); /* set two bits as 01 */
    UnlockMutex(m_Symlink_lock);
}

/* check whether the file system information received from prim client is consistent with files on disk */
BOOL CClientFileSysInfo::CheckConsistency(int checkTime)
{
/*printf("Want to CheckConsistency (Y/N)? ");
int ch = getchar();
if (ch == 'N')
return TRUE;*/

    if (m_fsbase == NULL || m_OptDirNum <= 0)
        return FALSE;

    CSysTime start_time, cur_time(0, 0), used_time(0, 0);
    int i, checkStatus, checkCont, checkIndex;
    int size, mode;
    BOOL isdir;
    char name[MAX_PATHLEN];

    int checkNum = 100;

    int total_IONum = (m_IOSize[0] + 1) * m_OptDirNum;
    int total_FileNum = m_TotalFileNum;
    int total_DirNum = m_TotalDirNum;
#ifndef WIN32
    int total_SymlinkNum = m_TotalSymlinkNum;
#endif

    for (i=0; i<checkNum; i++)
    {
        checkStatus = GetRandom() % 100;

#ifndef WIN32
        if (checkStatus < 25)
            checkCont = 1;      // for IO File
        else if (checkStatus < 50)
            checkCont = 2;      // for NonIO File
        else if (checkStatus < 75)
            checkCont = 3;      // for Dir
        else
            checkCont = 4;      // for Symlink
#else
        if (checkStatus < 34)
            checkCont = 1;      // for IO File
        else if (checkStatus < 67)
            checkCont = 2;      // for NonIO File
        else
            checkCont = 3;      // for Dir
#endif

        if (checkCont == 1)     // for IO File
        {
            while ( 1 )
            {
                checkIndex = GetRandom() % total_IONum;
                if (checkIndex / m_OptDirNum < m_IONum[checkIndex % m_OptDirNum])
                    break;
            }

            m_fsbase->GetIOFileName(name, m_OptDir[checkIndex % m_OptDirNum], checkIndex / m_OptDirNum);
            size = 0;
            if (CSafeFileOperation::Lookup(name, &size, NULL) != 1 || size != m_IOSize[checkIndex])
            {
                fprintf(stderr, "Error IO File: %s %d %d\n", name, size, m_IOSize[checkIndex]);
                return FALSE;
            }
        }

        if (checkCont == 2)     // for NonIO File
        {
            checkIndex = GetRandom() % total_FileNum;
            m_fsbase->GetNonIOFileName(name, m_OptDir[checkIndex % m_OptDirNum], checkIndex / m_OptDirNum);
            if (m_FileMode[checkIndex] & 0x0200)     // without this file
            {
                if (CSafeFileOperation::Lookup(name, NULL, NULL) != 2)
                {
                    fprintf(stderr, "Error NonIO File: %s shoule be deleted\n", name);
                    return FALSE;
                }
            }
            else                                    // with this file
            {
                if (CSafeFileOperation::Lookup(name, &size, &mode) != 1 || size != 0)
                {
                    fprintf(stderr, "Error NonIO File: %s shoule exist\n", name);
                    return FALSE;
                }
#ifndef WIN32
                isdir = ((mode & S_IFDIR) != 0);
#else
                isdir = ((((DWORD)mode) & FILE_ATTRIBUTE_DIRECTORY) != 0);
#endif
                if (isdir)
                {
                    fprintf(stderr, "Error NonIO File: %s shoule not be a dir\n", name);
                    return FALSE;
                }
                //if ((mode & 0x01FF) != (m_FileMode[checkIndex] & 0x01FF))
                //    return FALSE;
            }
        }

        if (checkCont == 3)     // for Dir
        {
            checkIndex = GetRandom() % total_DirNum;
            m_fsbase->GetDirName(name, m_OptDir[checkIndex % m_OptDirNum], checkIndex / m_OptDirNum);
            if (((m_DirMode[checkIndex/4] >> ((checkIndex%4) * 2)) & 0x01) == 0x01) // with this dir
            {
                if (CSafeFileOperation::Lookup(name, &size, &mode) != 1 || size != 0)
                {
                    fprintf(stderr, "Error Dir: %s shoule exist\n", name);
                    return FALSE;
                }
#ifndef WIN32
                isdir = ((mode & S_IFDIR) != 0);
#else
                isdir = ((((DWORD)mode) & FILE_ATTRIBUTE_DIRECTORY) != 0);
#endif
                if (! isdir)
                {
                    fprintf(stderr, "Error Dir: %s shoule be a dir\n", name);
                    return FALSE;
                }
            }
            else    // without this dir
            {
                if (CSafeFileOperation::Lookup(name, NULL, NULL) != 2)
                {
                    fprintf(stderr, "Error Dir: %s shoule be deleted\n", name);
                    return FALSE;
                }
            }
        }

#ifndef WIN32
        if (checkCont == 4)     // for Symlink
        {
            checkIndex = GetRandom() % total_SymlinkNum;
            m_fsbase->GetSymlinkName(name, m_OptDir[checkIndex % m_OptDirNum], checkIndex / m_OptDirNum);

            if (((m_SymlinkMode[checkIndex/4] >> ((checkIndex%4) * 2)) & 0x01) == 0x01) // with this symlink
            {
                if (CSafeFileOperation::Lookup(name, NULL, &mode) != 1)
                {
                    fprintf(stderr, "Error Symlink: %s shoule exist\n", name);
                    return FALSE;
                }
                if (! (mode & S_IFLNK))
                {
                    fprintf(stderr, "Error Symlink: %s shoule be symlink\n", name);
                    return FALSE;
                }
            }
            else    // without this symlink
            {
                if (CSafeFileOperation::Lookup(name, NULL, NULL) != 2)
                {
                    fprintf(stderr, "Error Symlink: %s shoule be deleted\n", name);
                    return FALSE;
                }
            }
        }
#endif

        cur_time.GetCurTime();
        used_time = cur_time - start_time;
        if (used_time.m_sec > checkTime)
            break;
    }

    return TRUE;
}
