
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysInfo.h
**
** DESCRIPTION: This is header file to define the class:
**              CClientFileSysInfo
**
**              This class mainly manage the file system information used
**              by the client.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
**  CLASS LIST:
**      CClientFileSysInfo
**
**
****************************************************************************/

#ifndef _CLIENTFILESYSINFO_H_
#define _CLIENTFILESYSINFO_H_

#include "Common.h"
#include "ClientFileSysBase.h"

#define SUBDIR_MAXMNUM                  997

class CClientFileSysInfo
{
private:
#ifndef WIN32
    pthread_mutex_t m_IO_lock;                  /* the lock of IO files structure */
    pthread_mutex_t m_NonIO_lock;               /* the lock of NonIO files structure */
    pthread_mutex_t m_Dir_lock;                 /* the lock of directories structure */
    pthread_mutex_t m_Symlink_lock;             /* the lock of symlinks structure */
#else
    HANDLE m_IO_lock;                           /* the lock of IO files structure */
    HANDLE m_NonIO_lock;                        /* the lock of NonIO files structure */
    HANDLE m_Dir_lock;                          /* the lock of directories structure */
    HANDLE m_Symlink_lock;                      /* the lock of symlinks structure */
#endif

    CClientFileSysBase *m_fsbase;               /* the base file system information */

    int m_IOSizeLevel;                          /* the level of file size for IO operation */

    int m_OptDirNum;                            /* the number of sub directory used by this client */
    int m_OptDir[SUBDIR_MAXMNUM];               /* sub directory name used by this client */

    /* variable definition for IO operation */
    int m_IONum[SUBDIR_MAXMNUM];                /* the number of IO file in each sub directory */
    int *m_IOSize;                              /* size of each IO file in each sub directory, the index of file i
                                                /* in subdir j is (i*m_OptDirNum)+j */
    int m_IOAcceNum;                            /* the number of file in working set */
    int *m_IOAcceSet;                           /* working set accessed by IO operations, store the fullname of the file */
                                                /* the fullname of the file i in subdir j is (i*m_OptDirNum)+j */
    int m_IOAcceGroupNum;                       /* the group number */
    int m_IOAccesGroupSize;                     /* the group size, some group may have one more file */
    int *m_IOAcceRange;                         /* the access range for each group in working set */
    int m_IOAcceMaxRange;                       /* the max range */
    BOOL m_IsBestFit;                           /* strategy to search a file in one group, BestFit or FirstFit */

    /* variable definition for NonIO operation */
    int m_TotalFileNum;                         /* the total number of all NonIO files, each remove operation will decrease */
                                                /* this number, each create operation will increase this number */
    int m_MinFileNum;                           /* the min number of file we may have */
    int m_MaxFileNum;                           /* the max number of file we may allocate */
    ushort *m_FileMode;                         /* mode of each NonIO file in each sub directory, 0x200 means file is deleted, 
                                                /* if the file is being used by an operation, the highest bit is set */
                                                /* the index of file i in subdir j is (i*m_OptDirNum)+j */
    int m_NewFile_Index;                        /* the index of the next new file */

    /* variable definition for directory operation */
    int m_TotalDirNum;                          /* the total number of directories , each rmdir operation will decrease */
                                                /* this number, each mkdir operation will increase this number */
    int m_MinDirNum;                            /* the min number of directory we may have */
    int m_MaxDirNum;                            /* the max number of directory we may allocate */
    uchar *m_DirMode;                           /* m_DirMode[index] (8bits) is the mode of 4 dirs, 2bits for each dir, 
                                                /* 1x means being used, 01 means dir exist, 00 means dir deleted */
    int m_NewDir_Index;                         /* the index of the next new directory */

    /* variable definition for symlink operation */
    int m_TotalSymlinkNum;                      /* the total number of symlinks , each rmsymlink operation will decrease */
                                                /* this number, each symlink operation will increase this number */
    int m_MinSymlinkNum;                        /* the min number of symlink we may have */
    int m_MaxSymlinkNum;                        /* the max number of symlink we may allocate */
    uchar *m_SymlinkMode;                       /* m_SymlinkMode[index] (8bits) is the mode of 4 symlinks, 2bits for each symlink,
                                                /* 1x means being used, 01 means symlink exist, 00 means symlink deleted */
    int m_NewSymlink_Index;                     /* the index of the next new symlink */

/* constructor and destructor */
public:
    CClientFileSysInfo(CClientFileSysBase *fsbase=NULL);
	~CClientFileSysInfo();

/* operations */
public:
    void InitSubDir(int num, int *subdir);
    void InitIOFileInfo(int *num, int **size);
    void InitNonIOFileInfo(int *num, ushort **mode);
    void InitDirInfo(int *num, uchar **mode);
    void InitSymlinkInfo(int *num, uchar **mode);
    void InitIOFileAcce();

    void CollSubDir(int &num, int *&subdir);
    void CollIOFileInfo(int *&num, int **&size);
    void CollNonIOFileInfo(int &totalnum, int *&num, ushort **&mode);
    void CollDirInfo(int &totalnum, int *&num, uchar **&mode);
    void CollSymlinkInfo(int &totalnum, int *&num, uchar **&mode);

    BOOL GetIOFileInfo(int req_size, char *name, int &handle, int &size);
    void SetIOFileInfo(int handle, int size);

    BOOL GetNonIOFileInfo(int option, char *name, int &handle, int &mode);  /* option: 0(access) 1(remove) 2(create) */
    void SetNonIOFileInfo(int option, int handle, int mode);                /* option: 0(access) 1(remove) 2(create) */

    BOOL GetDirInfo(int option, char *name, int &handle);                   /* option: 0(access) 1(remove) 2(create) */
    void SetDirInfo(int option, int handle);                                /* option: 0(access) 1(remove) 2(create) */

    BOOL GetSymlinkInfo(int option, char *name, int &handle);               /* option: 0(access) 1(remove) 2(create) */
    void SetSymlinkInfo(int option, int handle);                            /* option: 0(access) 1(remove) 2(create) */

    BOOL CheckConsistency(int checkTime);   // checkTime is in second
};

#endif

