
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientLoadGenPool.cpp
**
** DESCRIPTION: This class mainly provide functions to manage the threads
**              to generate load from client to file server.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientLoadGenPool::CClientLoadGenPool()
**     CClientLoadGenPool::~CClientLoadGenPool()
**
**
****************************************************************************/


#include "Common.h"
#include "ClientLoadGenPool.h"

/* define the only one LoadGenPool, it's a shared resource for all threads */
/*CClientLoadGenPool theLoadGenPool;*/

CClientLoadGenPool::CClientLoadGenPool(CClientLoadMgrThread *pLoadMgrThread)
{
    m_LoadMgrThread = pLoadMgrThread;

    m_LoadGenNum = 0;
    m_LoadGen = NULL;
    InitMutex(m_lock);
#ifndef WIN32
    pthread_cond_init(&m_event, NULL);
    m_IdleNum = 0;
#else
    m_hIdleSem = CreateSemaphore(NULL, 0, MAX_LOADGENNUM, NULL);
#endif
    m_IsIdle = NULL;
    m_NextIdleID = 0;
}

CClientLoadGenPool::CClientLoadGenPool(int num, CClientLoadGenThread **LoadGen, CClientLoadMgrThread *pLoadMgrThread)
{
    int i;

    m_LoadMgrThread = pLoadMgrThread;

    m_LoadGenNum = num;
    m_LoadGen = new CClientLoadGenThread * [num];
    for (i=0; i<num; i++)
        m_LoadGen[i] = LoadGen[i];

    InitMutex(m_lock);
#ifndef WIN32
    pthread_cond_init(&m_event, NULL);
    m_IdleNum = m_LoadGenNum;
#else
    m_hIdleSem = CreateSemaphore(NULL, m_LoadGenNum, MAX_LOADGENNUM, NULL);
#endif

    m_IsIdle = new BOOL [m_LoadGenNum];
    for (i=0; i<num; i++)
        m_IsIdle[i] = TRUE;
    m_NextIdleID = 0;
}

CClientLoadGenPool::~CClientLoadGenPool()
{
#ifdef WIN32
    if (m_lock != NULL)
        CloseHandle(m_lock);
    if (m_hIdleSem != NULL)
        CloseHandle(m_hIdleSem);
#endif

    if (m_LoadGen != NULL)
    {
        for (int i=0; i<m_LoadGenNum; i++)
        {
            m_LoadGen[i]->Stop();
            delete m_LoadGen[i];
        }
        delete [] m_LoadGen;
    }
    if (m_IsIdle != NULL)
        delete [] m_IsIdle;
}

int CClientLoadGenPool::GetLoadGenThreadNum()
{
    return m_LoadGenNum;
}

void CClientLoadGenPool::AddLoadGenThread(int num, CClientLoadGenThread **LoadGen)
{
    if (m_LoadGen != NULL)
        delete [] m_LoadGen;
    if (m_IsIdle != NULL)
        delete [] m_IsIdle;

    int i;
    m_LoadGenNum = num;
    m_LoadGen = new CClientLoadGenThread * [num];
    for (i=0; i<num; i++)
        m_LoadGen[i] = LoadGen[i];

#ifndef WIN32
    m_IdleNum = m_LoadGenNum;
#else
    ReleaseSemaphore(m_hIdleSem, m_LoadGenNum, NULL);
#endif

    m_IsIdle = new BOOL [m_LoadGenNum];
    for (i=0; i<num; i++)
        m_IsIdle[i] = TRUE;
    m_NextIdleID = 0;
}

int CClientLoadGenPool::GetLoadGenThreadID(CClientLoadGenThread *pLoadGen)
{
    int id;
    for (id=0; id<m_LoadGenNum; id++)
    {
        if (m_LoadGen[id] == pLoadGen)
            break;
    }
    return (id == m_LoadGenNum) ? -1 : id;
}

CClientLoadGenThread *CClientLoadGenPool::GetLoadGenThreadByID(int id)
{
    return m_LoadGen[id];
}

void CClientLoadGenPool::WairForAllIdle()
{
#ifndef WIN32
    LockMutex(m_lock);
    while (m_IdleNum < m_LoadGenNum)
        pthread_cond_wait(&m_event, &m_lock);
    UnlockMutex(m_lock);
#else
    for (int i=0; i<m_LoadGenNum; i++)
        WaitForSingleObject(m_hIdleSem, INFINITE);
    ReleaseSemaphore(m_hIdleSem, m_LoadGenNum, NULL);
#endif
}

CClientLoadGenThread *CClientLoadGenPool::GetIdleLoadGenThread()
{
#ifndef WIN32
    LockMutex(m_lock);
    while (m_IdleNum == 0)
        pthread_cond_wait(&m_event, &m_lock);
    m_IdleNum --;
#else
    //WaitForSingleObject(m_hIdleSem, INFINITE);
    //LockMutex(m_lock);

    int status;
    while ( 1 )
    {
        status = WaitForSingleObject(m_hIdleSem, 1000);
        if (status == WAIT_TIMEOUT)
        {
            printf("can't get idle thread\n");
            return NULL;
        }
        else
            break;
    }

    LockMutex(m_lock);

#endif

    int i = m_NextIdleID;
    /*printf("Pool: select thread from %d -- %d %d\n", i, m_IsIdle[0], m_IsIdle[1]);*/
    while (! m_IsIdle[i])
    {
        i ++;
        if (i == m_LoadGenNum)
            i = 0;
    }
    m_IsIdle[i] = FALSE;
    CClientLoadGenThread *pLoadGen = m_LoadGen[i];
    m_NextIdleID = i + 1;
    if (m_NextIdleID == m_LoadGenNum)
        m_NextIdleID = 0;
    /*printf("Pool: got thread %d (id=%d)\n", i, pLoadGen->m_id);*/

    UnlockMutex(m_lock);

    return pLoadGen;
}

void CClientLoadGenPool::SetIdleLoadGenThread(int id)
{
    LockMutex(m_lock);
    m_IsIdle[id] = TRUE;
#ifndef WIN32
    m_IdleNum ++;
    pthread_cond_signal(&m_event);
#else
    ReleaseSemaphore(m_hIdleSem, 1, NULL);
#endif
    UnlockMutex(m_lock);
}
