
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientLoadMgrThread.cpp
**
** DESCRIPTION: This class generates operation and picks a idle loadgen
**              thread to execute the operation.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientLoadMgrThread::CClientLoadMgrThread()
**     CClientLoadMgrThread::~CClientLoadMgrThread()
**     void CClientLoadMgrThread::run()
**
**
****************************************************************************/


#ifndef WIN32
#include <unistd.h>
#endif
#include "Common.h"
#include "ClientLoadMgrThread.h"
#include "ClientLoadGenPool.h"
#include "SocketClient.h"
#include "Packet.h"
#include "NetXDT.h"
#include "SysTime.h"
#include "Pumper.h"


CClientLoadMgrThread::CClientLoadMgrThread(CSocketClient *socket)
{
    m_socket = socket;

    m_LoadGenPool = new CClientLoadGenPool(this);
    m_fsinfoMgr = new CClientFileSysManager(this);
    m_opMgr = new CClientOperationManager(this);

    InitMutex(m_ops_lock);

#ifndef WIN32
    pthread_mutex_init(&m_ready_lock, NULL);
    pthread_cond_init(&m_ready_event, NULL);
    m_ready = FALSE;
#else
    m_ready_Sem = CreateSemaphore(NULL, 0, 1, NULL);
#endif

    InitMutex(m_break_lock);
    m_break = FALSE;

//    strcpy(m_LocalPath, "/LGFS/ffs1");
    m_mount_noac = TRUE;
}

CClientLoadMgrThread::~CClientLoadMgrThread()
{
    extern CPumper thePumper;

    if (m_socket != NULL)
        delete m_socket;
    if (m_LoadGenPool != NULL)
        delete m_LoadGenPool;
    if (m_fsinfoMgr != NULL)
        delete m_fsinfoMgr;
    if (m_opMgr != NULL)
        delete m_opMgr;

#ifdef WIN32
    if (m_ops_lock != NULL)
        CloseHandle(m_ops_lock);
    if (m_ready_Sem != NULL)
        CloseHandle(m_ready_Sem);
    if (m_break_lock != NULL)
        CloseHandle(m_break_lock);
#endif

    thePumper.DeleteThread(3, this);
}

CClientFileSysManager *CClientLoadMgrThread::GetFileSysManager()
{
    return m_fsinfoMgr;
}

CClientOperationManager *CClientLoadMgrThread::GetOperationManager()
{
    return m_opMgr;
}

int CClientLoadMgrThread::GetLoadID()
{
    return m_load_id;
}

int CClientLoadMgrThread::GetClientID()
{
    return m_client_id;
}

char *CClientLoadMgrThread::GetPrimName()
{
    return m_prim;
}

int CClientLoadMgrThread::GetLoad()
{
    return m_load;
}

void CClientLoadMgrThread::InitLoadGenThreadPool(int num)
{
    if (m_LoadGenPool == NULL)
        m_LoadGenPool = new CClientLoadGenPool(this);

    CClientLoadGenThread **loadgen = new CClientLoadGenThread * [num];
    for (int i=0; i<num; i++)
    {
        loadgen[i] = new CClientLoadGenThread(i, m_LoadGenPool);
        loadgen[i]->Start();
    }
    m_LoadGenPool->AddLoadGenThread(num, loadgen);
    delete [] loadgen;
}

void CClientLoadMgrThread::SetAutoRefreshed(BOOL auto_ref)
{
    m_auto_refreshed = auto_ref;
}

void CClientLoadMgrThread::SetOps(int ops)
{
    LockMutex(m_ops_lock);
    m_ops = ops;
    UnlockMutex(m_ops_lock);
}

int CClientLoadMgrThread::GetOps()
{
    LockMutex(m_ops_lock);
    int ops = m_ops;
    UnlockMutex(m_ops_lock);
    return ops;
}

void CClientLoadMgrThread::AddOps(int ops)
{
    LockMutex(m_ops_lock);
    m_ops += ops;
    UnlockMutex(m_ops_lock);
}

void CClientLoadMgrThread::ResetReady()
{
#ifndef WIN32
    LockMutex(m_ready_lock);
    m_ready = FALSE;
    UnlockMutex(m_ready_lock);
#endif
}

/* return TRUE: read, FALSE: timeout */
BOOL CClientLoadMgrThread::WaitForReady(int timeout)
{
#ifndef WIN32
    struct timespec to;
    to.tv_sec = time(NULL) + timeout / 1000;
    to.tv_nsec = (timeout % 1000) * 1000;
    LockMutex(m_ready_lock);
    BOOL ready = TRUE;
    while (! m_ready)
    {
        if (pthread_cond_timedwait(&m_ready_event, &m_ready_lock, &to) != 0)
        {
            ready = FALSE;
            break;
        }
    }
    m_ready = FALSE;
    UnlockMutex(m_ready_lock);
    return ready;
#else
    return WaitForSingleObject(m_ready_Sem, timeout) != WAIT_TIMEOUT;
#endif
}

void CClientLoadMgrThread::SetReady()
{
#ifndef WIN32
    LockMutex(m_ready_lock);
    m_ready = TRUE;
    pthread_cond_signal(&m_ready_event);
    UnlockMutex(m_ready_lock);
#else
    ReleaseSemaphore(m_ready_Sem, 1, NULL);
#endif
}

void CClientLoadMgrThread::ResetBreak()
{
    LockMutex(m_break_lock);
    m_break = 0;
    UnlockMutex(m_break_lock);
}

// 0: no break   1: break for shuffle   2: exit
int CClientLoadMgrThread::IsBreak()
{
    LockMutex(m_break_lock);
    int isBreak = m_break;
    UnlockMutex(m_break_lock);
    return isBreak;
}

// 0: no break   1: break for shuffle   2: exit
void CClientLoadMgrThread::SetBreak(int reason)
{
    LockMutex(m_break_lock);
    m_break = reason;
    UnlockMutex(m_break_lock);
}

void CClientLoadMgrThread::AddFileSysBase(CNet_FsSysBase *net_fsbase)
{
    char localPath[MAX_PATHLEN];

#ifndef WIN32
    strcpy(localPath, "");  /* will assign a path later */
#else
    sprintf(localPath, "\\\\%s\\%s", net_fsbase->ServerNetIF, net_fsbase->ShareName);

//DEBUG
//sprintf(localPath, "c:\\Stressor\\LGFS");
#endif

    CClientFileSysBase *fsbase = new CClientFileSysBase(net_fsbase->ServerName, net_fsbase->ServerNetIF,
                                                        net_fsbase->ServerPath,  localPath, 
                                                        net_fsbase->TotalSize, net_fsbase->TotalFNodeNum, 
                                                        net_fsbase->SubDirNum, net_fsbase->TotalIONum, 
                                                        net_fsbase->TotalFileNum, net_fsbase->TotalDirNum,
                                                        net_fsbase->TotalSymlinkNum);

    SFileSizeDist dist;
    dist.num = 10;
    dist.item[0].pcnt =  33;    dist.item[0].size =    1;    /* 33% - 1K */
    dist.item[1].pcnt =  54;    dist.item[1].size =    2;    /* 21% - 2K */
    dist.item[2].pcnt =  67;    dist.item[2].size =    4;    /* 13% - 4K */
    dist.item[3].pcnt =  77;    dist.item[3].size =    8;    /* 10% - 8K */
    dist.item[4].pcnt =  85;    dist.item[4].size =   16;    /*  8% - 16K */
    dist.item[5].pcnt =  90;    dist.item[5].size =   32;    /*  5% - 32K */
    dist.item[6].pcnt =  94;    dist.item[6].size =   64;    /*  4% - 64K */
	dist.item[7].pcnt =  97;    dist.item[7].size =  128;    /*  3% - 128K */
    dist.item[8].pcnt =  99;    dist.item[8].size =  256;    /*  2% - 256K */
    dist.item[9].pcnt = 100;    dist.item[9].size = 1024;    /*  1% - 1M */
    fsbase->SetFileSizeDist(&dist);

    if (! m_mount_noac)
        fsbase->SetCached();

    m_fsinfoMgr->AddFileSysBase(fsbase, 10);
}

void CClientLoadMgrThread::UpdateFileSysInfo(int index, CNet_FsSysInfo *net_fsinfo)
{
    CClientFileSysInfo *fsinfo = new CClientFileSysInfo (m_fsinfoMgr->GetFileSysBase(index));
    fsinfo->InitSubDir(net_fsinfo->OptDirNum, net_fsinfo->OptDir);
    fsinfo->InitIOFileInfo(net_fsinfo->IONum, net_fsinfo->IOSize);
    fsinfo->InitNonIOFileInfo(net_fsinfo->FileNum, net_fsinfo->FileMode);
    fsinfo->InitDirInfo(net_fsinfo->DirNum, net_fsinfo->DirMode);
    fsinfo->InitSymlinkInfo(net_fsinfo->SymlinkNum, net_fsinfo->SymlinkMode);
    fsinfo->InitIOFileAcce();
    m_fsinfoMgr->SetFileSysInfo(index, fsinfo);
}

BOOL CClientLoadMgrThread::RecvStartLoadFromPrim(CSocketClient *msg_sock)
{
    CPacket packet;
    char *buffer;
    int size;

    /* receive the load information from prim client firstly */
    if (! packet.Receive(m_socket))
        return FALSE;
    buffer = packet.GetData(&size);
    printf("got %d bytes\n", size);

    /* analyze the message */
    CNet_StartLoadOnClient net_startload;
    net_startload.xdt_trans(XDTOP_GET, buffer, size);

    m_load_id = net_startload.load_id;
    m_client_id = net_startload.client_id;
    strcpy(m_prim, net_startload.prim_name);
    strcpy(m_prim_ip, net_startload.prim_ip);
    strcpy(m_myname, net_startload.load_name);
    strcpy(m_myip, net_startload.load_ip);
    m_load = net_startload.opnum;
    m_stress_server = net_startload.stress_server;

    printf("load_id = %d\n", m_load_id);
    printf("client_id = %d\n", m_client_id);
    printf("prim = %s\n", m_prim);
    printf("load = %d\n", m_load);

    /* receive advanced setting */
    if (! packet.Receive(m_socket))
        return FALSE;
    buffer = packet.GetData(&size);
    CNet_LoadAdvance net_loadadv;
    net_loadadv.xdt_trans(XDTOP_GET, buffer, size);

    /* set thread number & op mix from advanced setting */
    InitLoadGenThreadPool(net_loadadv.thread_num);
    //m_opMsg->SetTotalTargetCalls(m_load * net_startload.run_time);
    m_opMgr->SetOpMix(net_loadadv.opmix_num, net_loadadv.opmix);

    m_mount_noac = net_loadadv.mount_noac;

    return TRUE;
}

BOOL CClientLoadMgrThread::RecvFSBaseInfoFromPrim(CSocketClient *msg_sock)
{
    CPacket packet;
    char *buffer;
    int size;

    printf("Receiving file system number ...\n");
    if (! packet.Receive(msg_sock))
        return FALSE;
    buffer = packet.GetData(&size);
    CNet_CommNum net_fsnum;
    net_fsnum.xdt_trans(XDTOP_GET, buffer, size);
    int fsnum = net_fsnum.num;
    printf("Got %d bytes, fsnum = %d\n", size, fsnum);

    printf("Receiving %d file system base information ...\n", fsnum);

    CNet_FsSysBase net_fsbase;
    int i;
    for (i=0; i<fsnum; i++)
    {
        if (! packet.Receive(msg_sock))
            return FALSE;
        buffer = packet.GetData(&size);
        printf("    File system %d received (%d bytes)\n", i, size);

        net_fsbase.xdt_trans(XDTOP_GET, buffer, size);
        printf("        ServerName = %s\n", net_fsbase.ServerName);
        printf("        ServerNetIF = %s\n", net_fsbase.ServerNetIF);
        printf("        ServerPath = %s\n", net_fsbase.ServerPath);
        printf("        ShareName = %s\n", net_fsbase.ShareName);

        AddFileSysBase(&net_fsbase);
    }

    printf("Mount %d file system ...\n", fsnum);
    CClientFileSysBase *fsbase;
    for (i=0; i<fsnum; i++)
    {
        fsbase = m_fsinfoMgr->GetFileSysBase(i);
        printf("        File system %d: %s %s\n", i+1, fsbase->GetServerNetIF(), fsbase->GetServerPath());
        fsbase->CheckAndMount();
    }
    printf("Finished\n");

    return TRUE;
}

BOOL CClientLoadMgrThread::RecvGoAheadFromPrim(CSocketClient *msg_sock, int timeout)
{
    CPacket packet;
    char *buffer;
    int size;

    if (! packet.Receive(msg_sock, timeout))
        return FALSE;
    buffer = packet.GetData(&size);
    CNet_CommNum net_signal;
    net_signal.xdt_trans(XDTOP_GET, buffer, size);
    return net_signal.num == 0x4321cdef;    /* check with this magic number */
}

BOOL CClientLoadMgrThread::RecvFileSysInfoFromPrim(CSocketClient *msg_sock)
{
    CPacket packet;
    char *buffer;
    int size;
    CNet_FsSysInfo net_fsinfo;

    int fsnum = m_fsinfoMgr->GetFileSysNum();
    printf("Receiving %d file system information ...\n", fsnum);
    for (int i=0; i<fsnum; i++)
    {
        if (! packet.Receive(msg_sock))
            return FALSE;
        buffer = packet.GetData(&size);
        net_fsinfo.xdt_trans(XDTOP_GET, buffer, size);
        UpdateFileSysInfo(i, &net_fsinfo);
        printf("    File system %d finished (%d bytes)\n", i, size);
    }

    return TRUE;
}

BOOL CClientLoadMgrThread::RecvBreakMessageFromPrim(CSocketClient *msg_sock)
{
    CPacket packet;
    char *buffer;
    int size;

    if (! packet.Receive(msg_sock))
        return FALSE;
    buffer = packet.GetData(&size);
    CNet_BreakMsg net_breakmsg;
    net_breakmsg.xdt_trans(XDTOP_GET, buffer, size);
    SetBreak(net_breakmsg.reason);
    return TRUE;
}

BOOL CClientLoadMgrThread::SendAckToPrim(int ack_code)
{
    /* send to local machine */
    //if (strcmp(m_myname, m_prim) == 0)
    {
    }

    CSocketClient socket;

    if (! socket.ConnectToServer(m_prim_ip, CLIENT_MAIN_PORT))
    {
        fprintf(stderr, "Cannot connect to primary client %s.\n", m_prim);
        return FALSE;
    }

    CPacket packet;
    char *buffer;
    int size;

    /* send the message type first */
    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_PUMPER_LOAD_ACK;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* then send the load id */
    CNet_LoadID net_loadid;
    net_loadid.id = m_load_id;
    net_loadid.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* then send the client id */
    CNet_ClientID net_clientid;
    net_clientid.id = m_client_id;
    net_clientid.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* then send the client id */
    CNet_CommNum net_ack;
    net_ack.num = ack_code;
    net_ack.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    return TRUE;
}

BOOL CClientLoadMgrThread::SendFileSysInfoToPrim()
{
    /* send to local machine */
    //if (strcmp(m_myname, m_prim) == 0)
    {
    }

    CSocketClient socket;

    if (! socket.ConnectToServer(m_prim_ip, CLIENT_MAIN_PORT))
    {
        fprintf(stderr, "Cannot connect to primary client %s.\n", m_prim);
        return FALSE;
    }

    CPacket packet;
    char *buffer;
    int size;

    /* send the message type first */
    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_PUMPER_LOAD_FSINFO;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* then send the load id */
    CNet_LoadID net_loadid;
    net_loadid.id = m_load_id;
    net_loadid.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* then send the client id */
    CNet_ClientID net_clientid;
    net_clientid.id = m_client_id;
    net_clientid.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* send the net_fssysinfo data */
    int i, j;
    CClientFileSysInfo *fsinfo;
    CNet_FsSysInfo net_fssysinfo;

    int fsnum = m_fsinfoMgr->GetFileSysNum();
    for (i=0; i<fsnum; i++)
    {
        fsinfo = m_fsinfoMgr->GetFileSysInfo(i);
        fsinfo->CollSubDir(net_fssysinfo.OptDirNum, net_fssysinfo.OptDir);
        fsinfo->CollIOFileInfo(net_fssysinfo.IONum, net_fssysinfo.IOSize);
        fsinfo->CollNonIOFileInfo(net_fssysinfo.TotalFileNum, net_fssysinfo.FileNum, net_fssysinfo.FileMode);
        fsinfo->CollDirInfo(net_fssysinfo.TotalDirNum, net_fssysinfo.DirNum, net_fssysinfo.DirMode);
        fsinfo->CollSymlinkInfo(net_fssysinfo.TotalSymlinkNum, net_fssysinfo.SymlinkNum, net_fssysinfo.SymlinkMode);

        net_fssysinfo.xdt_trans(XDTOP_PUT, buffer, size);

        /* delete the temp data */
        for (j=0; j<net_fssysinfo.OptDirNum; j++)
        {
            delete [] net_fssysinfo.IOSize[j];
            delete [] net_fssysinfo.FileMode[j];
            delete [] net_fssysinfo.DirMode[j];
            delete [] net_fssysinfo.SymlinkMode[j];
        }
        delete [] net_fssysinfo.OptDir;
        delete [] net_fssysinfo.IONum;
        delete [] net_fssysinfo.IOSize;
        delete [] net_fssysinfo.FileNum;
        delete [] net_fssysinfo.FileMode;
        delete [] net_fssysinfo.DirNum;
        delete [] net_fssysinfo.DirMode;
        delete [] net_fssysinfo.SymlinkNum;
        delete [] net_fssysinfo.SymlinkMode;

        /* the data is ready, now send it */
        packet.SetData(buffer, size);
        if (! packet.Send(&socket))
            return FALSE;
    }

    return TRUE;
}


#define RECORD_OPS_LEN      5


// return  1: break for shuffle   2: exit  3: timeout(exceed SHUFFLE_TIME)
int CClientLoadMgrThread::DoOp()
{
    int run_one_time = SHUFFLE_TIME + 5;

    if (m_stress_server)
//    if ( 1 )
    {
        CClientLoadGenThread *idle_thread;

        CSysTime t1, t2(0,0), t3(0,0);
        int prev_sec = 0;

        int ops;
        SetOps(0);

        int status = 0;

        while (status == 0)
        {
            /* select an idle thread to do actual op */
            idle_thread = m_LoadGenPool->GetIdleLoadGenThread();
if (idle_thread == NULL)
{
printf("%x: break status = %d\n", this, IsBreak());
continue;
}

/*printf("Thread %d: do op\n", idle_thread->m_id);
_SLEEP(1);
idle_thread->SetIdle();
continue;*/

            idle_thread->SetOpHostAndStart(this);

            t2.GetCurTime();
            t3 = t2 - t1;
            if (t3.m_sec > prev_sec)
            {
                ops = GetOps();
                printf("%03d secs %04d ops: %lf ops/sec\n", t3.m_sec, ops, (double)ops/(double)t3.m_sec);
                prev_sec = t3.m_sec;
            }

            if (t3.m_sec > run_one_time)
            {
                status = 3;
                break;
            }

            status = IsBreak();
        }

        ResetBreak();

        /* wait for all working threads finish their work */
        //_SLEEP(2);
        m_LoadGenPool->WairForAllIdle();

        return status;
    }

    else
    {
        CClientLoadGenThread *idle_thread;

        CSysTime start_time(0,0), cur_time(0,0), last_dis_time(0,0), last_check_time(0,0), run_time(0,0), dis_time(0,0), check_time(0,0);
        int prev_print_sec, period_msec, target_group_time_msec, real_group_time_msec;
        //double dis_time_sec;
        double run_time_sec;
        int check_time_msec, run_time_msec;
        int min_sleep_time_msec, sleep_time_msec;
        int target_load_per_sec, load_per_group;
        int ops, last_dis_ops, ops_group;

        int record_ops[RECORD_OPS_LEN];
        CSysTime record_time[RECORD_OPS_LEN];
        int total_record_num, next_record_index;

        total_record_num = next_record_index = 0;

        int status;

        SetOps(0);

        last_dis_ops = ops_group = 0;
        //last_check_ops = 0;

        target_load_per_sec = m_load;
        load_per_group = m_LoadGenPool->GetLoadGenThreadNum() / 2;

        period_msec = 1000;     /* check the ops/sec every 1 seconds */
        min_sleep_time_msec = 10;
        sleep_time_msec = 0;
        target_group_time_msec = 1000 * load_per_group / target_load_per_sec;

//printf("target group time = %d msec\n", target_group_time_msec);

        prev_print_sec = 0;

        start_time.GetCurTime();
        last_dis_time = start_time;
        last_check_time = start_time;

        status = 0;

        while (status == 0)
        {
            /* select an idle thread to do actual op */
            idle_thread = m_LoadGenPool->GetIdleLoadGenThread();
if (idle_thread == NULL)
{
printf("%x: break status = %d\n", this, IsBreak());
continue;
}

/*printf("Thread %d: do op\n", idle_thread->m_id);
_SLEEP(1);
idle_thread->SetIdle();
continue;*/

            idle_thread->SetOpHostAndStart(this);
            ops_group ++;

            cur_time.GetCurTime();
            run_time = cur_time - start_time;
            if (run_time.m_sec > prev_print_sec)
            {
                //dis_time = cur_time - last_dis_time;
                //last_dis_time = cur_time;
                //dis_time_sec = (double) dis_time.m_sec + dis_time.m_usec * 0.000001;

                run_time_sec = (double) run_time.m_sec + run_time.m_usec * 0.000001;

                ops = GetOps();
//                printf("%03d secs %04d ops: %lf ops/sec\n", run_time.m_sec, ops, (double)(ops - last_dis_ops)/dis_time_sec);
                printf("%03d secs %04d ops: %lf ops/sec\n", run_time.m_sec, ops, (double)ops/run_time_sec);
                last_dis_ops = ops;
                prev_print_sec = run_time.m_sec;
            }

            if (ops_group == load_per_group)
            {
                ops_group = 0;

                check_time = cur_time - last_check_time;
                check_time_msec = check_time.m_sec * 1000 + check_time.m_usec / 1000;
                if (check_time_msec > period_msec)     /* recaculate the sleep time every period */
                {
                    ops = GetOps();

                    last_check_time = cur_time;

                    //run_time_msec = run_time.m_sec * 1000 + run_time.m_usec / 1000;
                    //real_group_time_msec = run_time_msec * load_per_group / ops;

                    //real_group_time_msec = check_time_msec * load_per_group / (ops - last_check_ops);
                    //last_check_ops = ops;

                    if (total_record_num < RECORD_OPS_LEN)
                    {
                        record_ops[total_record_num] = ops;
                        record_time[total_record_num] = cur_time;
                        total_record_num ++;
                        run_time_msec = run_time.m_sec * 1000 + run_time.m_usec / 1000;
                        real_group_time_msec = run_time_msec * load_per_group / ops;
                    }
                    else
                    {
                        check_time = cur_time - record_time[next_record_index];
                        check_time_msec = check_time.m_sec * 1000 + check_time.m_usec / 1000;
                        real_group_time_msec = check_time_msec * load_per_group / (ops - record_ops[next_record_index]);
                        record_time[next_record_index] = cur_time;
                        record_ops[next_record_index] = ops;
                        next_record_index ++;
                        if (next_record_index == RECORD_OPS_LEN)
                            next_record_index = 0;
                    }


                    sleep_time_msec += target_group_time_msec - real_group_time_msec;
//printf("sleep_time = %d msec\n", sleep_time_msec);
                    //if (ops < target_ops_perc_sec * run_time_msec / 1000)
                    if (sleep_time_msec < 0)
                        sleep_time_msec = 0;
                }

                if (sleep_time_msec > 0)
                {
                    if (sleep_time_msec < min_sleep_time_msec)
                    {
                        if ((GetRandom() % 100) > 50)
                            MSecSleep(min_sleep_time_msec);
                    }
                    else
                    {
                        MSecSleep(sleep_time_msec);
                    }
                }
            }

            if (run_time.m_sec > run_one_time)
            {
                status = 3;
                break;
            }

            status = IsBreak();
        }

        ResetBreak();

        /* wait for all working threads finish their work */
        //_SLEEP(2);
        m_LoadGenPool->WairForAllIdle();

        return status;
    }
}

/*********************** entry-point of this thread ***********************/

void CClientLoadMgrThread::run()
{
    printf("Receive data of NET_MSGTYPE_PUMPER_START_LOAD ...\n");

    /* Receive from prim client: load parameter and advanced setting */
    fprintf(stderr, "CClientLoadMgrThread::run RecvStartLoadFromPrim()\n");
    if (! RecvStartLoadFromPrim(m_socket))
        return;

    /* Send ack to prim client to inform prim client that i'm ready and you can go ahead */
    fprintf(stderr, "CClientLoadMgrThread::run SendAckToPrim()\n");
    if (! SendAckToPrim(0))
        return;

    /* Receive from prim client: file system basic information */
    fprintf(stderr, "CClientLoadMgrThread::run RecvFSBaseInfoFromPrim()\n");
    if (! RecvFSBaseInfoFromPrim(m_socket))
        return;

    /* Check file system is mounted correctly */
    int ack = 0;
    int i;
    for (i=0; i<m_fsinfoMgr->GetFileSysNum(); i++)
    {
        if (! m_fsinfoMgr->GetFileSysBase(i)->IsMounted())
            ack |= (1 << i);
    }

    /* Send ack to prim client to inform prim client that i'm ready and you can go ahead,
       or tell prim client that some file system can't be mounted */
    /* If there are some file systems not mounted, can't continue */
    if (! SendAckToPrim(ack) || ack != 0)
        return;

    /* Receive from prim client: go ahead signal with magic number */
    if (! RecvGoAheadFromPrim(m_socket, 90000))
        return;

    /* Receive from prim client: file system information */
    if (! RecvFileSysInfoFromPrim(m_socket))
        return;

    /* Check if file system information is consistent with files on disk */
    ack = 0;
    for (i=0; i<m_fsinfoMgr->GetFileSysNum(); i++)
    {
        if (! m_fsinfoMgr->GetFileSysInfo(i)->CheckConsistency(15))
        {
            ack |= (1 << i);
            fprintf(stderr, "File system %d need to be refreshed\n", i);
        }
    }

    /* Send ack to prim client to inform prim client that i'm ready and you can go ahead,
       or tell prim client that some file system information is not consistent */
    /* If there are some file systems information not consistent, can't continue */
    if (! SendAckToPrim(ack) || ack != 0)
        return;

    /* Receive from prim client: go ahead signal with magic number */
    if (! RecvGoAheadFromPrim(m_socket, 60000))
        return;

    int status;
    while ( 1 )
    {
        /* Set this flag before each run loop */
        m_auto_refreshed = FALSE;

        /* Do operations until receive a break message */
        /* XXX */ /* need check Shuffle_time */
        status = DoOp();

        /* Timeout, too long not to receive break message from prim client */
        if (status == 3)
            break;

        /* Send updated file system information to prim client */
        fprintf(stderr, "Send updated file system information to prim client\n");
        if (! SendFileSysInfoToPrim())
        {
            fprintf(stderr, "Fail to send updated fs info to prim client\n");
            break;
        }
        fprintf(stderr, "Success to send updated fs info to prim client\n");

        /* Quit if received a quit message from prim client */
        if (status == 2)
            break;

        /* Wait for receiving new file system information for next run */
        /* SetReady() will be called by NET_MSGTYPE_PUMPER_FSINFO message receiver */
        /* XXX */ /* problem: if prim went into refresh step ??? here we just wait a long time (2 hours) */
        if (! WaitForReady(7200000))
            break;

        /* Check whether to be auto refreshed */
        if (m_auto_refreshed)
        {
            int fsnum = m_fsinfoMgr->GetFileSysNum();
            for (i=0; i<fsnum; i++)
                m_fsinfoMgr->GetFileSysBase(i)->CheckAndMount();
            m_auto_refreshed = FALSE;
        }

        /* Check if file system information is consistent with files on disk */
        ack = 0;
        for (i=0; i<m_fsinfoMgr->GetFileSysNum(); i++)
        {
            if (! m_fsinfoMgr->GetFileSysInfo(i)->CheckConsistency(15))
                ack |= (1 << i);
        }

        /* Send ack to prim client to inform prim client that i'm ready and you can go ahead,
           or tell prim client that some file system information is not consistent */
        /* If there are some file systems information not consistent, can't continue */
        if (! SendAckToPrim(ack) || ack != 0)
            break;

        /* Wait for go-ahead signal to keep running */
        /* SetReady() will be called by NET_MSGTYPE_PUMPER_GOAHEAD message receiver */
        if (! WaitForReady(30000))
            break;
    }
}

