
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientOperationManager.cpp
**
** DESCRIPTION: This class mainly manage how to generate an operation randomly.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientOperationManager::CClientOperationManager()
**     CClientOperationManager::~CClientOperationManager()
**     void CClientOperationManager::InitOpMap()
**     void CClientOperationManager::InitTargets() 
**     int CClientOperationManager::GetOpNum() 
**
****************************************************************************/


#include "Common.h"
#include "ClientOperationManager.h"
#include "ClientLoadMgrThread.h"
#include "ClientFileSysManager.h"

#include "Pumper.h"
extern CPumper thePumper;



CClientOperationManager::CClientOperationManager(CClientLoadMgrThread *lm_thread)
{
    m_LoadMgrThread = lm_thread;

    InitOpMap();

    InitTargets();
}

CClientOperationManager::~CClientOperationManager()
{

}

void CClientOperationManager::StaticInit()
{
    CClientOprtOverWrite::InitWriteBuffer();
    CClientOprtAppend::InitWriteBuffer();
}

void CClientOperationManager::SetOpMix(int opnum, int *opmix)
{
    for (int i=0; i<opnum; i++)
        m_OpMap[i].op_pcnt = opmix[i];
    InitTargets();
}

void CClientOperationManager::InitOpMap()
{
    for (int i = 0; i < MAX_OPNUM+1; i++)
    {
        strcpy(m_OpMap[i].op_name,  Ops_Pcnt[i].op_name);
        m_OpMap[i].op_pcnt = Ops_Pcnt[i].op_pcnt;
        m_OpMap[i].req_cnt = Ops_Pcnt[i].req_cnt;
        m_OpMap[i].req_pcnt = Ops_Pcnt[i].req_pcnt;
        m_OpMap[i].target_calls = Ops_Pcnt[i].target_calls;
        m_OpMap[i].target_reqs = Ops_Pcnt[i].target_reqs;
        m_OpMap[i].op = Ops_Pcnt[i].op;
    }
}

void CClientOperationManager::InitTargets()
{
    int		            call_target;		/* total number of calls to make */
    int		            req_target;		/* total number of reqs to make */
    int	                equal_mix;		/* equal mix of operations */
    int    	            slack;			/* calls leftover after % mix */
    int	            	randnum;		/* a random number */
    int             	i;			/* general use */
    double	            total_req_pcnt;
    double	            previous_pcnt;
    int                 clnt_req_load;
    int		            nops_used = 0;


    double avg_ops_per_read_req = m_LoadMgrThread->GetFileSysManager()->GetAvgOpsPerReadReq(); 
    double avg_ops_per_write_req = m_LoadMgrThread->GetFileSysManager()->GetAvgOpsPerWriteReq();
    int load_per_clnt = m_LoadMgrThread->GetLoad();

    /*
     * Compute number of target calls for each operation.
     * These are used to estimate the number of filehandles
     * that will be used for each type of operation.
     */
    call_target = m_OpMap[TOTAL].target_calls;
    m_OpMap[TOTAL].target_calls = 0;

    for (i = 0; i < MAX_OPNUM; i++) 
    {
	    m_OpMap[i].target_calls = (m_OpMap[i].op_pcnt * call_target) / 100;
	    m_OpMap[TOTAL].target_calls += m_OpMap[i].target_calls;
	    if (m_OpMap[i].op_pcnt != 0)
	        nops_used++;
    }

    /* Put left over calls into the heavier mix operations. */
    slack = call_target - m_OpMap[TOTAL].target_calls;
    equal_mix =  (100 / nops_used) / 2;
    while (slack > 0)
    {
	    randnum = GetRandom() % MAX_OPNUM;
	    if (m_OpMap[randnum].op_pcnt != 0 && m_OpMap[randnum].op_pcnt >= equal_mix) 
        {
	        m_OpMap[randnum].target_calls++;
	        m_OpMap[TOTAL].target_calls++;
	        slack--;
	    }
    }

    /*
     * compute request targets (based on suggestions from M. Molloy, HP)
     */

    /* compute total of target requests, based on weighted ops */
    total_req_pcnt = 0.0;
    for (i = 0; i < MAX_OPNUM ; i++) 
    {
	    switch (i) 
        {
	        case OPRT_READ:
	            total_req_pcnt += ((double) m_OpMap[i].op_pcnt) / avg_ops_per_read_req;
	            break;

	        case OPRT_OVERWRITE:
	            total_req_pcnt += ((double) m_OpMap[i].op_pcnt) / avg_ops_per_write_req;
	            break;

            case OPRT_APPEND:
	            total_req_pcnt += ((double) m_OpMap[i].op_pcnt) / avg_ops_per_write_req;
	            break;
            
            default:
	            total_req_pcnt += (double) m_OpMap[i].op_pcnt;
	            break;
	    }
    }

    /*
     * compute cumulative frequency distribution percentile for each op.
     * This code assumes that the NULLCALL does not generate multiple
     * OTW operations per request.
     */
    previous_pcnt = 0.0;
    for (i = 0; i < MAX_OPNUM; i++) 
    {
	    switch (i)
        {
	        case OPRT_READ:
	            m_OpMap[i].req_pcnt = previous_pcnt +
			                          (((100.0 * (double) m_OpMap[i].op_pcnt) / avg_ops_per_read_req) / total_req_pcnt);
	            break;

	        case OPRT_OVERWRITE:
	            m_OpMap[i].req_pcnt = previous_pcnt +
			                          (((100.0 * (double) m_OpMap[i].op_pcnt) / avg_ops_per_write_req) / total_req_pcnt);
	            break;

            case OPRT_APPEND:
	            m_OpMap[i].req_pcnt = previous_pcnt +
			                          (((100.0 * (double) m_OpMap[i].op_pcnt) / avg_ops_per_write_req) / total_req_pcnt);
	            break;

	        default:
	            m_OpMap[i].req_pcnt = previous_pcnt +
			                          ((100.0 * (double) m_OpMap[i].op_pcnt) / total_req_pcnt);
	            break;
	    }
	    previous_pcnt = m_OpMap[i].req_pcnt;
    }
    /* force last bucket to 100 */
    m_OpMap[MAX_OPNUM-1].req_pcnt = 100;

    /* compute the req load rate */
    clnt_req_load = (int) ((total_req_pcnt * load_per_clnt) / 100.0);

    /*
     * Compute number of target reqs for each operation.
     * These are used for debugging purposes.
     */
    req_target = (int) ((total_req_pcnt * m_OpMap[TOTAL].target_calls) / 100);
    m_OpMap[TOTAL].target_reqs = 0;

    previous_pcnt = 0.0;
    for (i = 0; i < MAX_OPNUM; i++)
    {
	    m_OpMap[i].target_reqs = 0;
	    if (m_OpMap[i].op_pcnt != 0) 
        {
	        m_OpMap[i].target_reqs = (int) (((m_OpMap[i].req_pcnt - previous_pcnt) * req_target) / 100);
	    }
	    m_OpMap[TOTAL].target_reqs += m_OpMap[i].target_reqs;
	    previous_pcnt = m_OpMap[i].req_pcnt;
    }

    /* Put left over reqs into the heavier mix operations. */
    slack = req_target - m_OpMap[TOTAL].target_reqs;
    equal_mix =  (100 / nops_used) / 2;
    while (slack > 0) 
    {
	    randnum = GetRandom() % MAX_OPNUM;
	    if (m_OpMap[randnum].target_reqs != 0 &&
					    m_OpMap[randnum].req_pcnt >= equal_mix) 
        {
	        m_OpMap[randnum].target_reqs++;
	        m_OpMap[TOTAL].target_reqs++;
	        slack--;
	    }
    }

}

void CClientOperationManager::GetTargetCall(int i, int &target_req, int &target_call)
{
    _ASSERT(i >= 0 && i < MAX_OPNUM);

    if ( i < TOTAL)
    {
        target_call = m_OpMap[i].target_calls;
        target_req  = m_OpMap[i].target_reqs;
    }
    else
    {
        target_req = 0;
        target_call = 0;
        for (int i = 0; i < MAX_OPNUM; i++)
        {
            target_call += m_OpMap[i].target_calls;
            target_req  += m_OpMap[i].target_reqs;
        }
    }
}


CClientOperation *CClientOperationManager::GetOp(CClientLoadMgrThread *host, char *opname)
{
    double ratio = double(GetRandom() % 10000) / 100.0;

//    int opnum;

    int opnum = thePumper.opnum;
    if (opnum == -1)
    {
        ratio = double(GetRandom() % 10000) / 100.0;
        for (opnum = 0; m_OpMap[opnum].req_pcnt <= ratio ; opnum++)
            ;

//XXX
        /*ratio = double(GetRandom() % 10000) / 100.0;
        if (ratio < 50)
            opnum = 0;
        else
            opnum = 1;*/

        //while ( 1 )
        //{
        //ratio = double(GetRandom() % 10000) / 100.0;
        //for (opnum = 0; m_OpMap[opnum].req_pcnt <= ratio ; opnum++)
        //    ;
//        if (opnum > 3 && opnum != 10 && opnum != 11 && opnum != 16)
        //if (opnum != 10 && opnum != 11)
        //    break;
        //}

/*        ratio = double(GetRandom() % 10000) / 100.0;
#ifndef WIN32
        if (ratio < 35)
            opnum = 0;
        else if (ratio < 70)
            opnum = 1;
        else if (ratio < 75)
            opnum = 4;
        else if (ratio < 80)
            opnum = 5;
        else if (ratio < 85)
            opnum = 9;
        else if (ratio < 90)
            opnum = 11;
        else if (ratio < 95)
            opnum = 15;
        else
            opnum = 16;
#else
        if (ratio < 40)
            opnum = 0;
        else if (ratio < 80)
            opnum = 1;
        else if (ratio < 85)
            opnum = 4;
        else if (ratio < 90)
            opnum = 5;
        else if (ratio < 95)
            opnum = 9;
        else
            opnum = 12;
#endif*/
    }

    strcpy(opname, m_OpMap[opnum].op_name);

    return m_OpMap[opnum].op->DupObj(host);
}



